"""Example real-time sleep scoring function."""

import numpy as np
import torch

from accusleepy.brain_state_set import BrainStateSet
from accusleepy.classification import get_device
from accusleepy.fileio import EMGFilter
from accusleepy.models import SSANN
from accusleepy.signal_processing import (
    create_eeg_emg_image,
    format_img,
    mixture_z_score_img,
)


def example_real_time_scoring_function(
    model: SSANN,
    eeg: np.ndarray,
    emg: np.ndarray,
    mixture_means: np.ndarray,
    mixture_sds: np.ndarray,
    sampling_rate: int | float,
    epoch_length: int | float,
    epochs_per_img: int,
    brain_state_set: BrainStateSet,
    emg_filter: EMGFilter,
) -> int:
    """Example function that could be used for real-time scoring

    This function demonstrates how you could use a model trained in
    "real-time" mode (current epoch on the right side of each image)
    to score incoming data. By passing a segment of EEG/EMG data
    into this function, the most recent epoch will be scored. For
    example, if the model expects 9 epochs worth of data and the
    epoch length is 5 seconds, you would pass in 45 seconds of data
    and would obtain the brain state of the most recent 5 seconds.

    Note:
    - The EEG and EMG signals must have length equal to
        sampling_rate * epoch_length * <number of epochs per image>.
    - The number of samples per epoch must be an integer.
    - This is just a demonstration, you should customize this for
        your application and there are probably ways to make it
        run faster.

    :param model: classification model
    :param eeg: EEG signal
    :param emg: EMG signal
    :param mixture_means: mixture means, for calibration
    :param mixture_sds: mixture standard deviations, for calibration
    :param sampling_rate: sampling rate, in Hz
    :param epoch_length: epoch length, in seconds
    :param epochs_per_img: number of epochs shown to the model at once
    :param brain_state_set: set of brain state options
    :param emg_filter: EMG filter parameters
    :return: brain state label
    """
    # prepare model
    # (this could be done outside the function)
    device = get_device()
    model = model.to(device)
    model.eval()

    # create and scale eeg+emg spectrogram
    img = create_eeg_emg_image(eeg, emg, sampling_rate, epoch_length, emg_filter)
    img, _ = mixture_z_score_img(
        img,
        mixture_means=mixture_means,
        mixture_sds=mixture_sds,
        brain_state_set=brain_state_set,
    )
    img = format_img(img=img, epochs_per_img=epochs_per_img, add_padding=False)

    # create dataset for inference
    images = torch.from_numpy(np.array([img.astype("float32")]))
    images = images[:, None, :, :]  # add channel
    images = images.to(device)

    # perform classification
    with torch.no_grad():
        outputs = model(images)
        _, predicted = torch.max(outputs, 1)

    label = int(brain_state_set.convert_class_to_digit(predicted.cpu().numpy())[0])
    return label



"""EEG/EMG signal processing, mixture z-scoring, and training image generation."""

import logging
import os
import warnings

import numpy as np
import pandas as pd
from PIL import Image
from tqdm import trange

from accusleepy.brain_state_set import BrainStateSet
from accusleepy.constants import (
    ABS_MAX_Z_SCORE,
    ANNOTATIONS_FILENAME,
    CALIBRATION_ANNOTATION_FILENAME,
    DEFAULT_MODEL_TYPE,
    DOWNSAMPLING_START_FREQ,
    EMG_COPIES,
    FILENAME_COL,
    LABEL_COL,
    MIN_EPOCHS_PER_STATE,
    MIN_WINDOW_LEN,
    SPECTROGRAM_UPPER_FREQ,
    UPPER_FREQ,
)
from accusleepy.fileio import EMGFilter, Recording, load_labels, load_recording
from accusleepy.multitaper import spectrogram

# note: scipy is lazily imported

logger = logging.getLogger(__name__)


def resample(
    eeg: np.ndarray,
    emg: np.ndarray,
    sampling_rate: int | float,
    epoch_length: int | float,
) -> tuple[np.ndarray, np.ndarray, float]:
    """Resample recording so that epochs contain equal numbers of samples

    If the number of samples per epoch is not an integer, epoch-level calculations
    are much more difficult. To avoid this, we can resample the EEG and EMG signals
    and adjust the sampling rate accordingly.

    :param eeg: EEG signal
    :param emg: EMG signal
    :param sampling_rate: original sampling rate, in Hz
    :param epoch_length: epoch length, in seconds
    :return: resampled EEG & EMG and updated sampling rate
    """
    samples_per_epoch = sampling_rate * epoch_length
    # round to the nearest integer rather than testing for exact equality:
    # epoch lengths that aren't exactly representable in binary (e.g., 2.2 s)
    # can make an integer samples-per-epoch look fractional, which would
    # otherwise trigger a needless (and sometimes large) resample
    target_samples_per_epoch = round(samples_per_epoch)
    if np.isclose(samples_per_epoch, target_samples_per_epoch):
        return eeg, emg, sampling_rate

    resampled = list()
    for arr in [eeg, emg]:
        x = np.arange(0, arr.size)
        x_new = np.linspace(
            0,
            arr.size - 1,
            round(arr.size * target_samples_per_epoch / samples_per_epoch),
        )
        resampled.append(np.interp(x_new, x, arr))

    eeg = resampled[0]
    emg = resampled[1]
    new_sampling_rate = target_samples_per_epoch / samples_per_epoch * sampling_rate
    return eeg, emg, new_sampling_rate


def standardize_signal_length(
    eeg: np.ndarray,
    emg: np.ndarray,
    sampling_rate: int | float,
    epoch_length: int | float,
) -> tuple[np.ndarray, np.ndarray]:
    """Truncate or pad EEG/EMG signals to have an integer number of epochs

    :param eeg: EEG signal
    :param emg: EMG signal
    :param sampling_rate: original sampling rate, in Hz
    :param epoch_length: epoch length, in seconds
    :return: EEG and EMG signals
    """
    # since resample() was called, this will be extremely close to an integer
    samples_per_epoch = round(sampling_rate * epoch_length)

    # pad the signal at the end in case we need more samples
    eeg = np.concatenate((eeg, np.ones(samples_per_epoch) * eeg[-1]))
    emg = np.concatenate((emg, np.ones(samples_per_epoch) * emg[-1]))
    padded_signal_length = eeg.size

    # count samples that don't fit in any epoch
    excess_samples = padded_signal_length % samples_per_epoch
    # we will definitely remove those
    last_index = padded_signal_length - excess_samples
    # and if the last epoch of real data had more than half of
    # its samples missing, delete it
    if excess_samples < samples_per_epoch / 2:
        last_index -= samples_per_epoch

    return eeg[:last_index], emg[:last_index]


def resample_and_standardize(
    eeg: np.ndarray,
    emg: np.ndarray,
    sampling_rate: int | float,
    epoch_length: int | float,
) -> tuple[np.ndarray, np.ndarray, float]:
    """Preprocess EEG and EMG signals

    Adjust the length and sampling rate of the EEG and EMG signals so that
    each epoch contains an integer number of samples and each recording
    contains an integer number of epochs.

    :param eeg: EEG signal
    :param emg: EMG signal
    :param sampling_rate: sampling rate, in Hz
    :param epoch_length: epoch length, in seconds
    :return: processed EEG & EMG signals, and the new sampling rate
    """
    eeg, emg, sampling_rate = resample(
        eeg=eeg, emg=emg, sampling_rate=sampling_rate, epoch_length=epoch_length
    )
    eeg, emg = standardize_signal_length(
        eeg=eeg, emg=emg, sampling_rate=sampling_rate, epoch_length=epoch_length
    )
    return eeg, emg, sampling_rate


def create_spectrogram(
    eeg: np.ndarray,
    sampling_rate: int | float,
    epoch_length: int | float,
    time_bandwidth: int = 2,
    n_tapers: int = 3,
) -> tuple[np.ndarray, np.ndarray]:
    """Create an EEG spectrogram image

    :param eeg: EEG signal
    :param sampling_rate: sampling rate, in Hz
    :param epoch_length: epoch length, in seconds
    :param time_bandwidth: time-half bandwidth product
    :param n_tapers: number of DPSS tapers to use
    :return: spectrogram and its frequency axis
    """
    window_length_sec = max(MIN_WINDOW_LEN, epoch_length)
    # pad the EEG signal so that the first spectrogram window is centered
    # on the first epoch
    # it's possible there's some jank here, if this isn't close to an integer
    pad_length = round(sampling_rate * (window_length_sec - epoch_length) / 2)
    padded_eeg = np.concatenate(
        [eeg[:pad_length][::-1], eeg, eeg[(len(eeg) - pad_length) :][::-1]]
    )

    spec, _, f = spectrogram(
        padded_eeg,
        sampling_rate,
        frequency_range=[0, SPECTROGRAM_UPPER_FREQ],
        time_bandwidth=time_bandwidth,
        num_tapers=n_tapers,
        window_params=[window_length_sec, epoch_length],
        min_nfft=0,
        detrend_opt="off",
        multiprocess=True,
        plot_on=False,
        return_fig=False,
        verbose=False,
    )

    # resample frequencies for consistency
    target_frequencies = np.arange(0, SPECTROGRAM_UPPER_FREQ, 1 / MIN_WINDOW_LEN)
    freq_idx = list()
    for i in target_frequencies:
        freq_idx.append(np.argmin(np.abs(f - i)))
    f = f[freq_idx]
    spec = spec[freq_idx, :]

    return spec, f


def get_emg_power(
    emg: np.ndarray,
    sampling_rate: int | float,
    epoch_length: int | float,
    emg_filter: EMGFilter,
) -> np.ndarray:
    """Calculate EMG power for each epoch

    This applies a 20-50 Hz bandpass filter to the EMG,  calculates the RMS
    in each epoch, and takes the log of the result.

    :param emg: EMG signal
    :param sampling_rate: sampling rate, in Hz
    :param epoch_length: epoch length, in seconds
    :param emg_filter: EMG filter parameters
    :return: EMG "power" for each epoch
    """
    from scipy.signal import butter, sosfiltfilt

    # use second-order sections rather than transfer-function ("ba") coefficients.
    # high-order Butterworth filters are numerically unstable in "ba" form, which
    # could produce all-NaN output for certain sampling rates (e.g., the rates
    # resample() yields for 891 Hz at 2.5 s epochs)
    sos = butter(
        N=emg_filter.order,
        Wn=[emg_filter.bp_lower, emg_filter.bp_upper],
        btype="bandpass",
        output="sos",
        fs=sampling_rate,
    )
    filtered = sosfiltfilt(sos, x=emg, padlen=int(np.ceil(sampling_rate)))

    # since resample() was called, this will be extremely close to an integer
    samples_per_epoch = round(sampling_rate * epoch_length)
    reshaped = np.reshape(
        filtered,
        [round(len(emg) / samples_per_epoch), samples_per_epoch],
    )
    rms = np.sqrt(np.mean(np.power(reshaped, 2), axis=1))
    # all-zero (e.g., absent) EMG gives rms == 0; log(0) = -inf is mapped to 0 below
    with np.errstate(divide="ignore"):
        log_rms = np.log(rms)
    log_rms[np.isinf(log_rms)] = 0
    return log_rms


def create_eeg_emg_image(
    eeg: np.ndarray,
    emg: np.ndarray,
    sampling_rate: int | float,
    epoch_length: int | float,
    emg_filter: EMGFilter,
) -> np.ndarray:
    """Stack EEG spectrogram and EMG power into an image

    This assumes that each epoch contains an integer number of samples and
    each recording contains an integer number of epochs. Note that a log
    transformation is applied to the spectrogram.

    :param eeg: EEG signal
    :param emg: EMG signal
    :param sampling_rate: sampling rate, in Hz
    :param epoch_length: epoch length, in seconds
    :param emg_filter: EMG filter parameters
    :return: combined EEG + EMG image for a recording
    """
    spec, f = create_spectrogram(eeg, sampling_rate, epoch_length)
    f_lower_idx = sum(f < DOWNSAMPLING_START_FREQ)
    f_upper_idx = sum(f < UPPER_FREQ)

    with np.errstate(divide="ignore"):
        modified_spectrogram = np.log(
            spec[
                np.concatenate(
                    [np.arange(0, f_lower_idx), np.arange(f_lower_idx, f_upper_idx, 2)]
                ),
                :,
            ]
        )
    # guard against zero-power bins (e.g., flatlined EEG segments) so that
    # -inf values don't propagate into the image
    if np.any(np.isneginf(modified_spectrogram)):
        finite_min = np.min(modified_spectrogram[np.isfinite(modified_spectrogram)])
        modified_spectrogram[np.isneginf(modified_spectrogram)] = finite_min

    emg_log_rms = get_emg_power(emg, sampling_rate, epoch_length, emg_filter)
    output = np.concatenate(
        [modified_spectrogram, np.tile(emg_log_rms, (EMG_COPIES, 1))]
    )
    return output


def get_mixture_values(
    img: np.ndarray, labels: np.ndarray, brain_state_set: BrainStateSet
) -> tuple[np.ndarray, np.ndarray]:
    """Compute weighted feature means and SDs for mixture z-scoring

    The outputs of this function can be used to standardize features
    extracted from all recordings from one subject under the same
    recording conditions. Note that labels must be in "class" format
    (i.e., integers between 0 and the number of scored states).

    :param img: combined EEG + EMG image - see create_eeg_emg_image()
    :param labels: brain state labels, in "class" format
    :param brain_state_set: set of brain state options
    :return: mixture means, mixture standard deviations
    """

    means = list()
    variances = list()
    mixture_weights = brain_state_set.mixture_weights

    # get feature means, variances by class
    for i in range(brain_state_set.n_classes):
        means.append(np.mean(img[:, labels == i], axis=1))
        variances.append(np.var(img[:, labels == i], axis=1))
    means = np.array(means)
    variances = np.array(variances)

    # mixture means are just weighted averages across classes
    mixture_means = means.T @ mixture_weights
    # mixture variance is given by the law of total variance
    mixture_sds = np.sqrt(
        variances.T @ mixture_weights
        + (
            (mixture_means - np.tile(mixture_means, (brain_state_set.n_classes, 1)))
            ** 2
        ).T
        @ mixture_weights
    )

    return mixture_means, mixture_sds


def mixture_z_score_img(
    img: np.ndarray,
    brain_state_set: BrainStateSet,
    labels: np.ndarray | None = None,
    mixture_means: np.ndarray | None = None,
    mixture_sds: np.ndarray | None = None,
) -> tuple[np.ndarray, bool]:
    """Perform mixture z-scoring on a combined EEG+EMG image

    If brain state labels are provided, they will be used to calculate
    mixture means and SDs. Otherwise, you must provide those inputs.
    Note that pixel values in the output are in the 0-1 range and will
    clip z-scores beyond ABS_MAX_Z_SCORE.

    :param img: combined EEG + EMG image - see create_eeg_emg_image()
    :param brain_state_set: set of brain state options
    :param labels: labels, in "class" format
    :param mixture_means: mixture means
    :param mixture_sds: mixture standard deviations
    :return: tuple of (z-scored image, whether zero-variance features were detected)
    """
    if labels is None and (mixture_means is None or mixture_sds is None):
        raise ValueError("must provide either labels or mixture means+SDs")
    if labels is not None and ((mixture_means is not None) ^ (mixture_sds is not None)):
        warnings.warn(
            "labels were given, mixture means / SDs will be ignored",
            stacklevel=2,
        )

    if labels is not None:
        mixture_means, mixture_sds = get_mixture_values(
            img=img, labels=labels, brain_state_set=brain_state_set
        )

    # replace zero SDs with epsilon to avoid division by zero
    # This can occur when a feature has no variance (e.g., no EMG signal)
    zero_sd_mask = mixture_sds == 0
    had_zero_variance = np.any(zero_sd_mask)
    if had_zero_variance:
        n_zero = np.sum(zero_sd_mask)
        logger.warning(
            "%s feature(s) have zero variance and will be mapped to neutral values",
            n_zero,
        )
        mixture_sds = mixture_sds.copy()
        mixture_sds[zero_sd_mask] = 1e-10

    img = ((img.T - mixture_means) / mixture_sds).T
    img = (img + ABS_MAX_Z_SCORE) / (2 * ABS_MAX_Z_SCORE)
    img = np.clip(img, 0, 1)

    return img, had_zero_variance


def format_img(img: np.ndarray, epochs_per_img: int, add_padding: bool) -> np.ndarray:
    """Adjust the format of an EEG+EMG image

    This function converts the values in a combined EEG+EMG image to uint8.
    This is a convenient format both for storing individual images as files,
    and for using the images as input to a classifier.
    This function also optionally adds new epochs to the beginning/end of the
    recording's image so that an image can be created for every epoch. For
    real-time scoring, padding should not be used.

    :param img: combined EEG + EMG image
    :param epochs_per_img: number of epochs in each individual image
    :param add_padding: whether to pad each side by (epochs_per_img - 1) / 2
    :return: formatted EEG + EMG image
    """
    # pad beginning and end
    if add_padding:
        pad_width = round((epochs_per_img - 1) / 2)
        img = np.concatenate(
            [
                np.tile(img[:, 0], (pad_width, 1)).T,
                img,
                np.tile(img[:, -1], (pad_width, 1)).T,
            ],
            axis=1,
        )

    # use 8-bit values
    img = np.clip(img * 255, 0, 255)
    img = img.astype(np.uint8)

    return img


def create_training_images(
    recordings: list[Recording],
    output_path: str,
    epoch_length: int | float,
    epochs_per_img: int,
    brain_state_set: BrainStateSet,
    model_type: str,
    calibration_fraction: float,
    emg_filter: EMGFilter,
) -> tuple[list[int], np.ndarray, bool]:
    """Create training dataset and calculate class balance

    This function creates images that can be used to train the
    SSANN model, as well as files that describe the training data
    (optionally split into training and calibration sets).
    It returns a list of recordings that could not be processed,
    the class balance of the usable training data, and a flag if any
    recordings had features with 0 variance.

    For each epoch, the model expects an image containing the
    EEG spectrogram and EMG power for several surrounding epochs.
    By default, the current epoch is located in the central column
    of pixels in each image. For real-time scoring applications,
    the current epoch is at the right edge of each image.

    :param recordings: list of recordings in the training set
    :param output_path: where to store training images
    :param epoch_length: epoch length, in seconds
    :param epochs_per_img: # number of epochs shown in each image
    :param brain_state_set: set of brain state options
    :param model_type: default or real-time
    :param calibration_fraction: fraction of training data to use for calibration
    :param emg_filter: EMG filter parameters
    :return: tuple of (failed recording names, training class balance, had zero-variance)
    """
    # recordings that had to be skipped
    failed_recordings = list()
    # image filenames for valid epochs
    filenames = list()
    # all valid labels from all valid recordings
    all_labels = list()
    # track if any recording had zero-variance features
    any_zero_variance = False
    # try to load each recording and create training images
    for i in trange(len(recordings)):
        recording = recordings[i]
        try:
            labels, _ = load_labels(recording.label_file)
        except Exception:
            logger.exception("Could not load labels for recording %s", recording.name)
            failed_recordings.append(recording.name)
            continue

        # Check that each scored brain state has sufficient observations
        # Ideally, we could use mixture means/SDs from another recording...
        insufficient_labels = False
        for brain_state in brain_state_set.brain_states:
            if brain_state.is_scored:
                count = np.sum(labels == brain_state.digit)
                if count < MIN_EPOCHS_PER_STATE:
                    logger.warning(
                        "Recording %s can't be used: insufficient labels for class '%s'",
                        recording.name,
                        brain_state.name,
                    )
                    failed_recordings.append(recording.name)
                    insufficient_labels = True
                    break
        if insufficient_labels:
            continue

        try:
            eeg, emg = load_recording(recording.recording_file)
            sampling_rate = recording.sampling_rate
            eeg, emg, sampling_rate = resample_and_standardize(
                eeg=eeg,
                emg=emg,
                sampling_rate=sampling_rate,
                epoch_length=epoch_length,
            )

            labels = brain_state_set.convert_digit_to_class(labels)
            img = create_eeg_emg_image(
                eeg, emg, sampling_rate, epoch_length, emg_filter
            )
            img, had_zero_variance = mixture_z_score_img(
                img=img, brain_state_set=brain_state_set, labels=labels
            )
            if had_zero_variance:
                any_zero_variance = True
            img = format_img(img=img, epochs_per_img=epochs_per_img, add_padding=True)

            # the model type determines which epochs are used in each image
            if model_type == DEFAULT_MODEL_TYPE:
                # here, j is the index of the current epoch in 'labels'
                # and the index of the leftmost epoch in 'img'
                for j in range(img.shape[1] - (epochs_per_img - 1)):
                    if labels[j] is None:
                        continue
                    im = img[:, j : (j + epochs_per_img)]
                    filename = f"recording_{recording.name}_{j}_{labels[j]}.png"
                    filenames.append(filename)
                    all_labels.append(labels[j])
                    Image.fromarray(im).save(os.path.join(output_path, filename))
            else:
                # here, j is the index of the current epoch in 'labels'
                # but we throw away a few epochs at the start since they
                # would require even more padding on the left side.
                one_side_padding = round((epochs_per_img - 1) / 2)
                for j in range(one_side_padding, len(labels)):
                    if labels[j] is None:
                        continue
                    im = img[:, (j - one_side_padding) : j + one_side_padding + 1]
                    filename = f"recording_{recording.name}_{j}_{labels[j]}.png"
                    filenames.append(filename)
                    all_labels.append(labels[j])
                    Image.fromarray(im).save(os.path.join(output_path, filename))

        except Exception:
            logger.exception(
                "Failed to create training images for recording %s", recording.name
            )
            failed_recordings.append(recording.name)

    annotations = pd.DataFrame({FILENAME_COL: filenames, LABEL_COL: all_labels})

    # split into training and calibration sets, if necessary
    if calibration_fraction > 0:
        calibration_set = annotations.sample(frac=calibration_fraction)
        training_set = annotations.drop(calibration_set.index)
        training_set.to_csv(
            os.path.join(output_path, ANNOTATIONS_FILENAME),
            index=False,
        )
        calibration_set.to_csv(
            os.path.join(output_path, CALIBRATION_ANNOTATION_FILENAME),
            index=False,
        )
        training_labels = training_set[LABEL_COL].values
    else:
        # annotation file contains info on all training images
        annotations.to_csv(
            os.path.join(output_path, ANNOTATIONS_FILENAME),
            index=False,
        )
        training_labels = np.array(all_labels)

    # compute class balance from training set
    class_counts = np.bincount(training_labels, minlength=brain_state_set.n_classes)
    training_class_balance = class_counts / class_counts.sum()
    logger.info("Training set class balance: %s", training_class_balance)

    return failed_recordings, training_class_balance, any_zero_variance



################################################################################
## Form generated from reading UI file 'viewer_window.ui'
##
## Created by: Qt User Interface Compiler version 6.10.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import QCoreApplication, QMetaObject, QSize, Qt
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import (
    QCheckBox,
    QFrame,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLayout,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QVBoxLayout,
)

from accusleepy.gui.mplwidget import MplWidget


class Ui_ViewerWindow:
    def setupUi(self, ViewerWindow):
        if not ViewerWindow.objectName():
            ViewerWindow.setObjectName("ViewerWindow")
        ViewerWindow.resize(1200, 700)
        sizePolicy = QSizePolicy(
            QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(ViewerWindow.sizePolicy().hasHeightForWidth())
        ViewerWindow.setSizePolicy(sizePolicy)
        ViewerWindow.setStyleSheet("background-color: white;")
        self.horizontalLayout = QHBoxLayout(ViewerWindow)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.all_plots = QVBoxLayout()
        self.all_plots.setSpacing(1)
        self.all_plots.setObjectName("all_plots")
        self.all_plots.setSizeConstraint(QLayout.SizeConstraint.SetMinimumSize)
        self.upperfigure = MplWidget(ViewerWindow)
        self.upperfigure.setObjectName("upperfigure")
        sizePolicy.setHeightForWidth(self.upperfigure.sizePolicy().hasHeightForWidth())
        self.upperfigure.setSizePolicy(sizePolicy)
        self.upperfigure.setAutoFillBackground(False)

        self.all_plots.addWidget(self.upperfigure)

        self.line = QFrame(ViewerWindow)
        self.line.setObjectName("line")
        self.line.setLineWidth(2)
        self.line.setFrameShape(QFrame.Shape.HLine)
        self.line.setFrameShadow(QFrame.Shadow.Sunken)

        self.all_plots.addWidget(self.line)

        self.lowerfigure = MplWidget(ViewerWindow)
        self.lowerfigure.setObjectName("lowerfigure")
        sizePolicy.setHeightForWidth(self.lowerfigure.sizePolicy().hasHeightForWidth())
        self.lowerfigure.setSizePolicy(sizePolicy)
        self.lowerfigure.setAutoFillBackground(False)

        self.all_plots.addWidget(self.lowerfigure)

        self.all_plots.setStretch(0, 50)
        self.all_plots.setStretch(1, 1)
        self.all_plots.setStretch(2, 50)

        self.horizontalLayout.addLayout(self.all_plots)

        self.all_controls = QVBoxLayout()
        self.all_controls.setSpacing(20)
        self.all_controls.setObjectName("all_controls")
        self.all_controls.setSizeConstraint(QLayout.SizeConstraint.SetDefaultConstraint)
        self.top_plot_buttons = QVBoxLayout()
        self.top_plot_buttons.setObjectName("top_plot_buttons")
        self.topcontroltopspacer = QSpacerItem(
            5, 10, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred
        )

        self.top_plot_buttons.addItem(self.topcontroltopspacer)

        self.save_help_buttons = QHBoxLayout()
        self.save_help_buttons.setSpacing(20)
        self.save_help_buttons.setObjectName("save_help_buttons")
        self.save_help_buttons.setSizeConstraint(
            QLayout.SizeConstraint.SetDefaultConstraint
        )
        self.savebutton = QPushButton(ViewerWindow)
        self.savebutton.setObjectName("savebutton")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.savebutton.sizePolicy().hasHeightForWidth())
        self.savebutton.setSizePolicy(sizePolicy1)
        self.savebutton.setStyleSheet("background-color: rgb(237, 241, 241);")
        icon = QIcon()
        icon.addFile(":/icons/save.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.savebutton.setIcon(icon)
        self.savebutton.setIconSize(QSize(28, 28))
        self.savebutton.setAutoRepeat(True)

        self.save_help_buttons.addWidget(self.savebutton)

        self.helpbutton = QPushButton(ViewerWindow)
        self.helpbutton.setObjectName("helpbutton")
        sizePolicy1.setHeightForWidth(self.helpbutton.sizePolicy().hasHeightForWidth())
        self.helpbutton.setSizePolicy(sizePolicy1)
        self.helpbutton.setStyleSheet("background-color: rgb(237, 241, 241);")
        icon1 = QIcon()
        icon1.addFile(
            ":/icons/question.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off
        )
        self.helpbutton.setIcon(icon1)
        self.helpbutton.setIconSize(QSize(28, 28))
        self.helpbutton.setAutoRepeat(True)

        self.save_help_buttons.addWidget(self.helpbutton)

        self.top_plot_buttons.addLayout(self.save_help_buttons)

        self.topcontrolmidspacer = QSpacerItem(
            5, 5, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred
        )

        self.top_plot_buttons.addItem(self.topcontrolmidspacer)

        self.zoom_and_brightness = QVBoxLayout()
        self.zoom_and_brightness.setObjectName("zoom_and_brightness")
        self.verticalSpacer_6 = QSpacerItem(
            5, 5, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred
        )

        self.zoom_and_brightness.addItem(self.verticalSpacer_6)

        self.brightness_buttons = QHBoxLayout()
        self.brightness_buttons.setObjectName("brightness_buttons")
        self.specbrighter = QPushButton(ViewerWindow)
        self.specbrighter.setObjectName("specbrighter")
        sizePolicy1.setHeightForWidth(
            self.specbrighter.sizePolicy().hasHeightForWidth()
        )
        self.specbrighter.setSizePolicy(sizePolicy1)
        self.specbrighter.setStyleSheet("background-color: rgb(237, 241, 241);")
        icon2 = QIcon()
        icon2.addFile(
            ":/icons/brightness_up.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off
        )
        self.specbrighter.setIcon(icon2)
        self.specbrighter.setIconSize(QSize(24, 24))
        self.specbrighter.setAutoRepeat(True)

        self.brightness_buttons.addWidget(self.specbrighter)

        self.specdimmer = QPushButton(ViewerWindow)
        self.specdimmer.setObjectName("specdimmer")
        sizePolicy1.setHeightForWidth(self.specdimmer.sizePolicy().hasHeightForWidth())
        self.specdimmer.setSizePolicy(sizePolicy1)
        self.specdimmer.setStyleSheet("background-color: rgb(237, 241, 241);")
        icon3 = QIcon()
        icon3.addFile(
            ":/icons/brightness_down.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off
        )
        self.specdimmer.setIcon(icon3)
        self.specdimmer.setIconSize(QSize(24, 24))
        self.specdimmer.setAutoRepeat(True)

        self.brightness_buttons.addWidget(self.specdimmer)

        self.zoom_and_brightness.addLayout(self.brightness_buttons)

        self.zoom_buttons = QHBoxLayout()
        self.zoom_buttons.setObjectName("zoom_buttons")
        self.zoom_buttons.setSizeConstraint(QLayout.SizeConstraint.SetDefaultConstraint)
        self.zoom_buttons.setContentsMargins(2, -1, 2, -1)
        self.xzoomin = QPushButton(ViewerWindow)
        self.xzoomin.setObjectName("xzoomin")
        sizePolicy1.setHeightForWidth(self.xzoomin.sizePolicy().hasHeightForWidth())
        self.xzoomin.setSizePolicy(sizePolicy1)
        self.xzoomin.setStyleSheet("background-color: rgb(237, 241, 241);")
        icon4 = QIcon()
        icon4.addFile(
            ":/icons/zoom_in.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off
        )
        self.xzoomin.setIcon(icon4)
        self.xzoomin.setIconSize(QSize(20, 20))
        self.xzoomin.setAutoRepeat(True)

        self.zoom_buttons.addWidget(self.xzoomin)

        self.xzoomout = QPushButton(ViewerWindow)
        self.xzoomout.setObjectName("xzoomout")
        sizePolicy1.setHeightForWidth(self.xzoomout.sizePolicy().hasHeightForWidth())
        self.xzoomout.setSizePolicy(sizePolicy1)
        self.xzoomout.setStyleSheet("background-color: rgb(237, 241, 241);")
        icon5 = QIcon()
        icon5.addFile(
            ":/icons/zoom_out.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off
        )
        self.xzoomout.setIcon(icon5)
        self.xzoomout.setIconSize(QSize(20, 20))
        self.xzoomout.setAutoRepeat(True)

        self.zoom_buttons.addWidget(self.xzoomout)

        self.xzoomreset = QPushButton(ViewerWindow)
        self.xzoomreset.setObjectName("xzoomreset")
        sizePolicy1.setHeightForWidth(self.xzoomreset.sizePolicy().hasHeightForWidth())
        self.xzoomreset.setSizePolicy(sizePolicy1)
        self.xzoomreset.setStyleSheet("background-color: rgb(237, 241, 241);")
        icon6 = QIcon()
        icon6.addFile(":/icons/home.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.xzoomreset.setIcon(icon6)
        self.xzoomreset.setIconSize(QSize(20, 20))
        self.xzoomreset.setAutoRepeat(True)

        self.zoom_buttons.addWidget(self.xzoomreset)

        self.zoom_and_brightness.addLayout(self.zoom_buttons)

        self.verticalSpacer_7 = QSpacerItem(
            5, 5, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred
        )

        self.zoom_and_brightness.addItem(self.verticalSpacer_7)

        self.top_plot_buttons.addLayout(self.zoom_and_brightness)

        self.topcontrolbottomspacer = QSpacerItem(
            5, 5, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred
        )

        self.top_plot_buttons.addItem(self.topcontrolbottomspacer)

        self.top_plot_buttons.setStretch(0, 2)
        self.top_plot_buttons.setStretch(1, 5)
        self.top_plot_buttons.setStretch(2, 5)
        self.top_plot_buttons.setStretch(3, 12)
        self.top_plot_buttons.setStretch(4, 12)

        self.all_controls.addLayout(self.top_plot_buttons)

        self.allcontrolmidspacer = QSpacerItem(
            5, 5, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred
        )

        self.all_controls.addItem(self.allcontrolmidspacer)

        self.bottom_plot_buttons = QVBoxLayout()
        self.bottom_plot_buttons.setObjectName("bottom_plot_buttons")
        self.verticalSpacer_3 = QSpacerItem(
            5, 5, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred
        )

        self.bottom_plot_buttons.addItem(self.verticalSpacer_3)

        self.eeg_emg_buttons = QVBoxLayout()
        self.eeg_emg_buttons.setSpacing(2)
        self.eeg_emg_buttons.setObjectName("eeg_emg_buttons")
        self.eegbuttons = QGridLayout()
        self.eegbuttons.setObjectName("eegbuttons")
        self.eegbuttons.setHorizontalSpacing(10)
        self.eegbuttons.setVerticalSpacing(20)
        self.eegshiftup = QPushButton(ViewerWindow)
        self.eegshiftup.setObjectName("eegshiftup")
        sizePolicy1.setHeightForWidth(self.eegshiftup.sizePolicy().hasHeightForWidth())
        self.eegshiftup.setSizePolicy(sizePolicy1)
        self.eegshiftup.setStyleSheet("background-color: rgb(237, 241, 241);")
        icon7 = QIcon()
        icon7.addFile(
            ":/icons/double_up_arrow.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off
        )
        self.eegshiftup.setIcon(icon7)
        self.eegshiftup.setIconSize(QSize(20, 20))
        self.eegshiftup.setAutoRepeat(True)

        self.eegbuttons.addWidget(self.eegshiftup, 0, 1, 1, 1)

        self.eegzoomout = QPushButton(ViewerWindow)
        self.eegzoomout.setObjectName("eegzoomout")
        sizePolicy1.setHeightForWidth(self.eegzoomout.sizePolicy().hasHeightForWidth())
        self.eegzoomout.setSizePolicy(sizePolicy1)
        self.eegzoomout.setStyleSheet("background-color: rgb(237, 241, 241);")
        self.eegzoomout.setIcon(icon5)
        self.eegzoomout.setIconSize(QSize(20, 20))
        self.eegzoomout.setAutoRepeat(True)

        self.eegbuttons.addWidget(self.eegzoomout, 1, 0, 1, 1)

        self.eegshiftdown = QPushButton(ViewerWindow)
        self.eegshiftdown.setObjectName("eegshiftdown")
        sizePolicy1.setHeightForWidth(
            self.eegshiftdown.sizePolicy().hasHeightForWidth()
        )
        self.eegshiftdown.setSizePolicy(sizePolicy1)
        self.eegshiftdown.setStyleSheet("background-color: rgb(237, 241, 241);")
        icon8 = QIcon()
        icon8.addFile(
            ":/icons/double_down_arrow.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off
        )
        self.eegshiftdown.setIcon(icon8)
        self.eegshiftdown.setIconSize(QSize(20, 20))
        self.eegshiftdown.setAutoRepeat(True)

        self.eegbuttons.addWidget(self.eegshiftdown, 1, 1, 1, 1)

        self.eegzoomin = QPushButton(ViewerWindow)
        self.eegzoomin.setObjectName("eegzoomin")
        sizePolicy1.setHeightForWidth(self.eegzoomin.sizePolicy().hasHeightForWidth())
        self.eegzoomin.setSizePolicy(sizePolicy1)
        self.eegzoomin.setStyleSheet("background-color: rgb(237, 241, 241);")
        self.eegzoomin.setIcon(icon4)
        self.eegzoomin.setIconSize(QSize(20, 20))
        self.eegzoomin.setAutoRepeat(True)

        self.eegbuttons.addWidget(self.eegzoomin, 0, 0, 1, 1)

        self.eeg_emg_buttons.addLayout(self.eegbuttons)

        self.verticalSpacer_8 = QSpacerItem(
            5, 5, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred
        )

        self.eeg_emg_buttons.addItem(self.verticalSpacer_8)

        self.reset_tip_label = QLabel(ViewerWindow)
        self.reset_tip_label.setObjectName("reset_tip_label")
        sizePolicy1.setHeightForWidth(
            self.reset_tip_label.sizePolicy().hasHeightForWidth()
        )
        self.reset_tip_label.setSizePolicy(sizePolicy1)

        self.eeg_emg_buttons.addWidget(self.reset_tip_label)

        self.verticalSpacer_5 = QSpacerItem(
            5, 5, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred
        )

        self.eeg_emg_buttons.addItem(self.verticalSpacer_5)

        self.emgbuttons = QGridLayout()
        self.emgbuttons.setObjectName("emgbuttons")
        self.emgbuttons.setHorizontalSpacing(10)
        self.emgbuttons.setVerticalSpacing(20)
        self.emgzoomin = QPushButton(ViewerWindow)
        self.emgzoomin.setObjectName("emgzoomin")
        sizePolicy1.setHeightForWidth(self.emgzoomin.sizePolicy().hasHeightForWidth())
        self.emgzoomin.setSizePolicy(sizePolicy1)
        self.emgzoomin.setStyleSheet("background-color: rgb(237, 241, 241);")
        self.emgzoomin.setIcon(icon4)
        self.emgzoomin.setIconSize(QSize(20, 20))
        self.emgzoomin.setAutoRepeat(True)

        self.emgbuttons.addWidget(self.emgzoomin, 0, 0, 1, 1)

        self.emgshiftdown = QPushButton(ViewerWindow)
        self.emgshiftdown.setObjectName("emgshiftdown")
        sizePolicy1.setHeightForWidth(
            self.emgshiftdown.sizePolicy().hasHeightForWidth()
        )
        self.emgshiftdown.setSizePolicy(sizePolicy1)
        self.emgshiftdown.setStyleSheet("background-color: rgb(237, 241, 241);")
        self.emgshiftdown.setIcon(icon8)
        self.emgshiftdown.setIconSize(QSize(20, 20))
        self.emgshiftdown.setAutoRepeat(True)

        self.emgbuttons.addWidget(self.emgshiftdown, 1, 1, 1, 1)

        self.emgshiftup = QPushButton(ViewerWindow)
        self.emgshiftup.setObjectName("emgshiftup")
        sizePolicy1.setHeightForWidth(self.emgshiftup.sizePolicy().hasHeightForWidth())
        self.emgshiftup.setSizePolicy(sizePolicy1)
        self.emgshiftup.setStyleSheet("background-color: rgb(237, 241, 241);")
        self.emgshiftup.setIcon(icon7)
        self.emgshiftup.setIconSize(QSize(20, 20))
        self.emgshiftup.setAutoRepeat(True)

        self.emgbuttons.addWidget(self.emgshiftup, 0, 1, 1, 1)

        self.emgzoomout = QPushButton(ViewerWindow)
        self.emgzoomout.setObjectName("emgzoomout")
        sizePolicy1.setHeightForWidth(self.emgzoomout.sizePolicy().hasHeightForWidth())
        self.emgzoomout.setSizePolicy(sizePolicy1)
        self.emgzoomout.setStyleSheet("background-color: rgb(237, 241, 241);")
        self.emgzoomout.setIcon(icon5)
        self.emgzoomout.setIconSize(QSize(20, 20))
        self.emgzoomout.setAutoRepeat(True)

        self.emgbuttons.addWidget(self.emgzoomout, 1, 0, 1, 1)

        self.eeg_emg_buttons.addLayout(self.emgbuttons)

        self.eeg_emg_buttons.setStretch(0, 10)
        self.eeg_emg_buttons.setStretch(1, 1)
        self.eeg_emg_buttons.setStretch(2, 1)
        self.eeg_emg_buttons.setStretch(3, 1)
        self.eeg_emg_buttons.setStretch(4, 10)

        self.bottom_plot_buttons.addLayout(self.eeg_emg_buttons)

        self.verticalSpacer = QSpacerItem(
            5, 5, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred
        )

        self.bottom_plot_buttons.addItem(self.verticalSpacer)

        self.autoscroll_layout = QVBoxLayout()
        self.autoscroll_layout.setObjectName("autoscroll_layout")
        self.autoscroll_layout.setContentsMargins(30, -1, 20, -1)
        self.autoscroll = QCheckBox(ViewerWindow)
        self.autoscroll.setObjectName("autoscroll")
        sizePolicy1.setHeightForWidth(self.autoscroll.sizePolicy().hasHeightForWidth())
        self.autoscroll.setSizePolicy(sizePolicy1)
        self.autoscroll.setStyleSheet("background-color: none;")

        self.autoscroll_layout.addWidget(self.autoscroll)

        self.bottom_plot_buttons.addLayout(self.autoscroll_layout)

        self.verticalSpacer_2 = QSpacerItem(
            5, 5, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred
        )

        self.bottom_plot_buttons.addItem(self.verticalSpacer_2)

        self.epoch_controls = QGridLayout()
        self.epoch_controls.setObjectName("epoch_controls")
        self.epoch_controls.setHorizontalSpacing(2)
        self.epoch_controls.setVerticalSpacing(10)
        self.shownepochslabel = QLabel(ViewerWindow)
        self.shownepochslabel.setObjectName("shownepochslabel")
        sizePolicy2 = QSizePolicy(
            QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Preferred
        )
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(
            self.shownepochslabel.sizePolicy().hasHeightForWidth()
        )
        self.shownepochslabel.setSizePolicy(sizePolicy2)
        self.shownepochslabel.setAlignment(
            Qt.AlignmentFlag.AlignLeading
            | Qt.AlignmentFlag.AlignLeft
            | Qt.AlignmentFlag.AlignVCenter
        )

        self.epoch_controls.addWidget(self.shownepochslabel, 0, 1, 1, 1)

        self.epochsword = QLabel(ViewerWindow)
        self.epochsword.setObjectName("epochsword")
        sizePolicy1.setHeightForWidth(self.epochsword.sizePolicy().hasHeightForWidth())
        self.epochsword.setSizePolicy(sizePolicy1)

        self.epoch_controls.addWidget(self.epochsword, 0, 0, 1, 1)

        self.shownepochsplus = QPushButton(ViewerWindow)
        self.shownepochsplus.setObjectName("shownepochsplus")
        sizePolicy1.setHeightForWidth(
            self.shownepochsplus.sizePolicy().hasHeightForWidth()
        )
        self.shownepochsplus.setSizePolicy(sizePolicy1)
        self.shownepochsplus.setStyleSheet("background-color: rgb(237, 241, 241);")
        icon9 = QIcon()
        icon9.addFile(
            ":/icons/up_arrow.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off
        )
        self.shownepochsplus.setIcon(icon9)
        self.shownepochsplus.setIconSize(QSize(20, 20))
        self.shownepochsplus.setAutoRepeat(False)

        self.epoch_controls.addWidget(self.shownepochsplus, 1, 1, 1, 1)

        self.shownepochsminus = QPushButton(ViewerWindow)
        self.shownepochsminus.setObjectName("shownepochsminus")
        sizePolicy1.setHeightForWidth(
            self.shownepochsminus.sizePolicy().hasHeightForWidth()
        )
        self.shownepochsminus.setSizePolicy(sizePolicy1)
        self.shownepochsminus.setStyleSheet("background-color: rgb(237, 241, 241);")
        icon10 = QIcon()
        icon10.addFile(
            ":/icons/down_arrow.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off
        )
        self.shownepochsminus.setIcon(icon10)
        self.shownepochsminus.setIconSize(QSize(20, 20))
        self.shownepochsminus.setAutoRepeat(False)

        self.epoch_controls.addWidget(self.shownepochsminus, 1, 0, 1, 1)

        self.epoch_controls.setColumnStretch(0, 1)
        self.epoch_controls.setColumnStretch(1, 1)

        self.bottom_plot_buttons.addLayout(self.epoch_controls)

        self.verticalSpacer_4 = QSpacerItem(
            5, 5, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed
        )

        self.bottom_plot_buttons.addItem(self.verticalSpacer_4)

        self.bottom_plot_buttons.setStretch(0, 1)
        self.bottom_plot_buttons.setStretch(1, 6)
        self.bottom_plot_buttons.setStretch(2, 1)
        self.bottom_plot_buttons.setStretch(3, 3)
        self.bottom_plot_buttons.setStretch(4, 1)
        self.bottom_plot_buttons.setStretch(5, 4)
        self.bottom_plot_buttons.setStretch(6, 1)

        self.all_controls.addLayout(self.bottom_plot_buttons)

        self.all_controls.setStretch(0, 50)
        self.all_controls.setStretch(1, 1)
        self.all_controls.setStretch(2, 50)

        self.horizontalLayout.addLayout(self.all_controls)

        self.horizontalLayout.setStretch(0, 20)
        self.horizontalLayout.setStretch(1, 1)

        self.retranslateUi(ViewerWindow)

        QMetaObject.connectSlotsByName(ViewerWindow)

    # setupUi

    def retranslateUi(self, ViewerWindow):
        # if QT_CONFIG(tooltip)
        self.savebutton.setToolTip(
            QCoreApplication.translate("ViewerWindow", "Save labels (Ctrl+S)", None)
        )
        # endif // QT_CONFIG(tooltip)
        self.savebutton.setText("")
        # if QT_CONFIG(tooltip)
        self.helpbutton.setToolTip(
            QCoreApplication.translate("ViewerWindow", "User manual", None)
        )
        # endif // QT_CONFIG(tooltip)
        self.helpbutton.setText("")
        # if QT_CONFIG(tooltip)
        self.specbrighter.setToolTip(
            QCoreApplication.translate("ViewerWindow", "Increase brightness", None)
        )
        # endif // QT_CONFIG(tooltip)
        self.specbrighter.setText("")
        # if QT_CONFIG(tooltip)
        self.specdimmer.setToolTip(
            QCoreApplication.translate("ViewerWindow", "Decrease brightness", None)
        )
        # endif // QT_CONFIG(tooltip)
        self.specdimmer.setText("")
        # if QT_CONFIG(tooltip)
        self.xzoomin.setToolTip(
            QCoreApplication.translate("ViewerWindow", "Zoom in (+)", None)
        )
        # endif // QT_CONFIG(tooltip)
        self.xzoomin.setText("")
        # if QT_CONFIG(tooltip)
        self.xzoomout.setToolTip(
            QCoreApplication.translate("ViewerWindow", "Zoom out (-)", None)
        )
        # endif // QT_CONFIG(tooltip)
        self.xzoomout.setText("")
        # if QT_CONFIG(tooltip)
        self.xzoomreset.setToolTip(
            QCoreApplication.translate("ViewerWindow", "Reset zoom", None)
        )
        # endif // QT_CONFIG(tooltip)
        self.xzoomreset.setText("")
        # if QT_CONFIG(tooltip)
        self.eegshiftup.setToolTip(
            QCoreApplication.translate("ViewerWindow", "Shift signal up", None)
        )
        # endif // QT_CONFIG(tooltip)
        self.eegshiftup.setText("")
        # if QT_CONFIG(tooltip)
        self.eegzoomout.setToolTip(
            QCoreApplication.translate("ViewerWindow", "Scale signal down", None)
        )
        # endif // QT_CONFIG(tooltip)
        self.eegzoomout.setText("")
        # if QT_CONFIG(tooltip)
        self.eegshiftdown.setToolTip(
            QCoreApplication.translate("ViewerWindow", "Shift signal down", None)
        )
        # endif // QT_CONFIG(tooltip)
        self.eegshiftdown.setText("")
        # if QT_CONFIG(tooltip)
        self.eegzoomin.setToolTip(
            QCoreApplication.translate("ViewerWindow", "Scale signal up", None)
        )
        # endif // QT_CONFIG(tooltip)
        self.eegzoomin.setText("")
        self.reset_tip_label.setText(
            QCoreApplication.translate("ViewerWindow", "double-click to reset", None)
        )
        self.emgzoomin.setText("")
        self.emgshiftdown.setText("")
        self.emgshiftup.setText("")
        self.emgzoomout.setText("")
        # if QT_CONFIG(tooltip)
        self.autoscroll.setToolTip(
            QCoreApplication.translate(
                "ViewerWindow", "Step forward when setting brain state", None
            )
        )
        # endif // QT_CONFIG(tooltip)
        self.autoscroll.setText(
            QCoreApplication.translate("ViewerWindow", "Auto scroll", None)
        )
        self.shownepochslabel.setText(
            QCoreApplication.translate("ViewerWindow", "5", None)
        )
        self.epochsword.setText(
            QCoreApplication.translate("ViewerWindow", "Epochs:", None)
        )
        self.shownepochsplus.setText("")
        self.shownepochsminus.setText("")
        pass

    # retranslateUi



"""File I/O for recordings, labels, calibration data, and configuration."""

import json
import os
import shutil
from dataclasses import dataclass
from importlib.metadata import PackageNotFoundError, version
from importlib.resources import files

import numpy as np
import pandas as pd
from platformdirs import user_config_dir

import accusleepy.constants as c
from accusleepy.brain_state_set import BRAIN_STATES_KEY, BrainState, BrainStateSet


@dataclass
class EMGFilter:
    """Convenience class for a EMG filter parameters"""

    order: int  # filter order
    bp_lower: int | float  # lower bandpass frequency
    bp_upper: int | float  # upper bandpass frequency


@dataclass
class Hyperparameters:
    """Convenience class for model training hyperparameters"""

    batch_size: int
    learning_rate: float
    momentum: float
    training_epochs: int


@dataclass
class AccuSleePyConfig:
    """AccuSleePy configuration settings"""

    brain_state_set: BrainStateSet
    default_epoch_length: int | float
    overwrite_setting: bool
    save_confidence_setting: bool
    min_bout_length: int | float
    emg_filter: EMGFilter
    hyperparameters: Hyperparameters
    epochs_to_show: int
    autoscroll_state: bool
    delete_training_images: bool


@dataclass
class Recording:
    """Store information about a recording"""

    name: int = 1  # name to show in the GUI
    recording_file: str = ""  # path to recording file
    label_file: str = ""  # path to label file
    calibration_file: str = ""  # path to calibration file
    sampling_rate: int | float = 0.0  # sampling rate, in Hz


def load_calibration_file(filename: str) -> tuple[np.ndarray, np.ndarray]:
    """Load a calibration file

    :param filename: filename
    :return: mixture means and SDs
    """
    df = pd.read_csv(filename)
    mixture_means = df[c.MIXTURE_MEAN_COL].values
    mixture_sds = df[c.MIXTURE_SD_COL].values
    return mixture_means, mixture_sds


def load_csv_or_parquet(filename: str) -> pd.DataFrame:
    """Load a csv or parquet file as a dataframe

    :param filename: filename
    :return: dataframe of file contents
    """
    extension = os.path.splitext(filename)[1]
    if extension == ".csv":
        df = pd.read_csv(filename)
    elif extension == ".parquet":
        df = pd.read_parquet(filename)
    else:
        raise ValueError("file must be csv or parquet")
    return df


def load_recording(filename: str) -> tuple[np.ndarray, np.ndarray]:
    """Load recording of EEG and EMG time series data

    :param filename: filename
    :return: arrays of EEG and EMG data
    """
    df = load_csv_or_parquet(filename)
    eeg = df[c.EEG_COL].values
    emg = df[c.EMG_COL].values
    return eeg, emg


def load_labels(filename: str) -> tuple[np.ndarray, np.ndarray | None]:
    """Load file of brain state labels and confidence scores

    :param filename: filename
    :return: array of brain state labels and, optionally, array of confidence scores
    """
    df = load_csv_or_parquet(filename)
    if c.CONFIDENCE_SCORE_COL in df.columns:
        return df[c.BRAIN_STATE_COL].values, df[c.CONFIDENCE_SCORE_COL].values
    else:
        return df[c.BRAIN_STATE_COL].values, None


def save_labels(
    labels: np.ndarray, filename: str, confidence_scores: np.ndarray | None = None
) -> None:
    """Save brain state labels to file

    :param labels: brain state labels
    :param filename: filename
    :param confidence_scores: optional confidence scores
    """
    if confidence_scores is not None:
        pd.DataFrame(
            {c.BRAIN_STATE_COL: labels, c.CONFIDENCE_SCORE_COL: confidence_scores}
        ).to_csv(filename, index=False)
    else:
        pd.DataFrame({c.BRAIN_STATE_COL: labels}).to_csv(filename, index=False)


def _get_user_config_path() -> str:
    """Return the path to the user's config file in the platform config directory."""
    return os.path.join(user_config_dir("accusleepy"), c.USER_CONFIG_FILE)


def _get_default_config_path() -> str:
    """Return the path to the bundled default config file."""
    return str(files("accusleepy").joinpath(c.DEFAULT_CONFIG_FILE))


def load_config() -> AccuSleePyConfig:
    """Load configuration file with brain state options

    :return: AccuSleePyConfig containing the following:
        set of brain state options,
        default epoch length,
        default overwrite setting,
        default confidence score output setting,
        default minimum bout length,
        EMG filter parameters,
        model training hyperparameters,
        default epochs to show for manual scoring,
        default autoscroll state for manual scoring,
        setting to delete training images automatically
    """
    user_config = _get_user_config_path()
    if not os.path.exists(user_config):
        os.makedirs(os.path.dirname(user_config), exist_ok=True)
        shutil.copy2(_get_default_config_path(), user_config)
    with open(user_config) as f:
        data = json.load(f)

    return AccuSleePyConfig(
        brain_state_set=BrainStateSet(
            [BrainState(**b) for b in data[BRAIN_STATES_KEY]], c.UNDEFINED_LABEL
        ),
        default_epoch_length=data[c.DEFAULT_EPOCH_LENGTH_KEY],
        overwrite_setting=data.get(
            c.DEFAULT_OVERWRITE_KEY, c.DEFAULT_OVERWRITE_SETTING
        ),
        save_confidence_setting=data.get(
            c.DEFAULT_CONFIDENCE_SETTING_KEY, c.DEFAULT_CONFIDENCE_SETTING
        ),
        min_bout_length=data.get(
            c.DEFAULT_MIN_BOUT_LENGTH_KEY, c.DEFAULT_MIN_BOUT_LENGTH
        ),
        emg_filter=EMGFilter(
            **data.get(
                c.EMG_FILTER_KEY,
                {
                    "order": c.DEFAULT_EMG_FILTER_ORDER,
                    "bp_lower": c.DEFAULT_EMG_BP_LOWER,
                    "bp_upper": c.DEFAULT_EMG_BP_UPPER,
                },
            )
        ),
        hyperparameters=Hyperparameters(
            **data.get(
                c.HYPERPARAMETERS_KEY,
                {
                    "batch_size": c.DEFAULT_BATCH_SIZE,
                    "learning_rate": c.DEFAULT_LEARNING_RATE,
                    "momentum": c.DEFAULT_MOMENTUM,
                    "training_epochs": c.DEFAULT_TRAINING_EPOCHS,
                },
            )
        ),
        epochs_to_show=data.get(c.EPOCHS_TO_SHOW_KEY, c.DEFAULT_EPOCHS_TO_SHOW),
        autoscroll_state=data.get(c.AUTOSCROLL_KEY, c.DEFAULT_AUTOSCROLL_STATE),
        delete_training_images=data.get(
            c.DELETE_TRAINING_IMAGES_KEY, c.DEFAULT_DELETE_TRAINING_IMAGES_STATE
        ),
    )


def save_config(
    brain_state_set: BrainStateSet,
    default_epoch_length: int | float,
    overwrite_setting: bool,
    save_confidence_setting: bool,
    min_bout_length: int | float,
    emg_filter: EMGFilter,
    hyperparameters: Hyperparameters,
    epochs_to_show: int,
    autoscroll_state: bool,
    delete_training_images: bool,
) -> None:
    """Save configuration of brain state options to json file

    :param brain_state_set: set of brain state options
    :param default_epoch_length: default epoch length
    :param save_confidence_setting: default setting for
        saving confidence scores
    :param emg_filter: EMG filter parameters
    :param min_bout_length: default minimum bout length
    :param overwrite_setting: default setting for overwriting
        existing labels
    :param hyperparameters: model training hyperparameters
    :param epochs_to_show: default epochs to show for manual scoring,
    :param autoscroll_state: default autoscroll state for manual scoring
    :param delete_training_images: whether to automatically delete images
        created for model training once training is complete
    """
    output_dict = brain_state_set.to_output_dict()
    output_dict.update({c.DEFAULT_EPOCH_LENGTH_KEY: default_epoch_length})
    output_dict.update({c.DEFAULT_OVERWRITE_KEY: overwrite_setting})
    output_dict.update({c.DEFAULT_CONFIDENCE_SETTING_KEY: save_confidence_setting})
    output_dict.update({c.DEFAULT_MIN_BOUT_LENGTH_KEY: min_bout_length})
    output_dict.update({c.EMG_FILTER_KEY: emg_filter.__dict__})
    output_dict.update({c.HYPERPARAMETERS_KEY: hyperparameters.__dict__})
    output_dict.update({c.EPOCHS_TO_SHOW_KEY: epochs_to_show})
    output_dict.update({c.AUTOSCROLL_KEY: autoscroll_state})
    output_dict.update({c.DELETE_TRAINING_IMAGES_KEY: delete_training_images})
    user_config = _get_user_config_path()
    os.makedirs(os.path.dirname(user_config), exist_ok=True)
    with open(user_config, "w") as f:
        json.dump(output_dict, f, indent=4)
        f.write("\n")


def load_recording_list(filename: str) -> list[Recording]:
    """Load list of recordings from file

    :param filename: filename of list of recordings
    :return: list of recordings
    """
    with open(filename) as f:
        data = json.load(f)
    recording_list = [Recording(**r) for r in data[c.RECORDING_LIST_NAME]]
    for i, r in enumerate(recording_list):
        r.name = i + 1
    return recording_list


def save_recording_list(filename: str, recordings: list[Recording]) -> None:
    """Save list of recordings to file

    :param filename: where to save the list
    :param recordings: list of recordings to export
    """
    recording_dict = {
        c.RECORDING_LIST_NAME: [
            {
                "recording_file": r.recording_file,
                "label_file": r.label_file,
                "calibration_file": r.calibration_file,
                "sampling_rate": r.sampling_rate,
            }
            for r in recordings
        ]
    }
    with open(filename, "w") as f:
        json.dump(recording_dict, f, indent=4)


def get_version() -> str:
    """Get AccuSleePy package version

    :return: AccuSleePy package version
    """
    try:
        return version("accusleepy")
    except PackageNotFoundError:
        return ""



"""AccuSleePy main window.

Icon source: Arkinasi, https://www.flaticon.com/authors/arkinasi
"""

import logging
import os
import sys
from dataclasses import dataclass
from functools import partial

import numpy as np
from PySide6.QtCore import (
    QEvent,
    QKeyCombination,
    QObject,
    QRect,
    Qt,
    QUrl,
)
from PySide6.QtGui import QKeySequence, QShortcut
from PySide6.QtWidgets import (
    QApplication,
    QLabel,
    QMainWindow,
    QTextBrowser,
    QVBoxLayout,
    QWidget,
)

from accusleepy.brain_state_set import BRAIN_STATES_KEY
from accusleepy.constants import (
    CALIBRATION_FILE_TYPE,
    DEFAULT_MODEL_TYPE,
    LABEL_FILE_TYPE,
    MESSAGE_BOX_MAX_DEPTH,
    MODEL_FILE_TYPE,
    REAL_TIME_MODEL_TYPE,
    RECORDING_FILE_TYPES,
    RECORDING_LIST_FILE_TYPE,
    UNDEFINED_LABEL,
)
from accusleepy.fileio import (
    get_version,
    load_config,
    load_labels,
    load_recording,
)
from accusleepy.gui.dialogs import select_existing_file, select_save_location
from accusleepy.gui.manual_scoring import ManualScoringWindow
from accusleepy.gui.primary_window import Ui_PrimaryWindow
from accusleepy.gui.recording_manager import RecordingListManager
from accusleepy.gui.settings_widget import SettingsWidget
from accusleepy.services import (
    LoadedModel,
    TrainingService,
    check_single_file_inputs,
    create_calibration,
    score_recording_list,
)
from accusleepy.signal_processing import resample_and_standardize
from accusleepy.validation import check_config_consistency, validate_and_correct_labels

logger = logging.getLogger(__name__)

# on Windows, prevent dark mode from changing the visual style
if os.name == "nt":
    sys.argv += ["-platform", "windows:darkmode=0"]


# relative path to user manual
MAIN_GUIDE_FILE = os.path.normpath(r"text/main_guide.md")


@dataclass
class TrainingSettings:
    """Settings for training a new model"""

    epochs_per_img: int = 9
    model_type: str = DEFAULT_MODEL_TYPE
    calibrate: bool = True


@dataclass
class ScoringSettings:
    """Settings for scoring a recording"""

    only_overwrite_undefined: bool
    save_confidence_scores: bool
    min_bout_length: int | float


class AccuSleepWindow(QMainWindow):
    """AccuSleePy primary window"""

    def __init__(self):
        super().__init__()

        # initialize the UI
        self.ui = Ui_PrimaryWindow()
        self.ui.setupUi(self)
        self.setWindowTitle("AccuSleePy")

        # Load configuration
        loaded_config = load_config()

        # Apply default values from the configuration
        self.epoch_length = loaded_config.default_epoch_length
        self.scoring = ScoringSettings(
            only_overwrite_undefined=loaded_config.overwrite_setting,
            save_confidence_scores=loaded_config.save_confidence_setting,
            min_bout_length=loaded_config.min_bout_length,
        )

        # Initialize settings tab (manages Settings tab UI and saved config values)
        self.config = SettingsWidget(ui=self.ui, config=loaded_config, parent=self)

        # initialize info about the recordings, classification data / settings
        self.ui.epoch_length_input.setValue(self.epoch_length)
        self.ui.overwritecheckbox.setChecked(self.scoring.only_overwrite_undefined)
        self.ui.save_confidence_checkbox.setChecked(self.scoring.save_confidence_scores)
        self.ui.bout_length_input.setValue(self.scoring.min_bout_length)

        # loaded classification model and its metadata
        self.loaded_model = LoadedModel()

        # settings for training new models
        self.training = TrainingSettings()

        # set up the list of recordings
        self.recording_manager = RecordingListManager(
            self.ui.recording_list_widget, parent=self
        )

        # messages to display
        self.messages = []

        # display current version
        self.ui.version_label.setText(f"v{get_version()}")

        # user input: keyboard shortcuts
        keypress_quit = QShortcut(
            QKeySequence(QKeyCombination(Qt.Modifier.CTRL, Qt.Key.Key_W)),
            self,
        )
        keypress_quit.activated.connect(self.close)

        # user input: button presses
        self.ui.add_button.clicked.connect(self.add_recording)
        self.ui.remove_button.clicked.connect(self.remove_recording)
        self.ui.recording_list_widget.currentRowChanged.connect(self.select_recording)
        self.ui.sampling_rate_input.valueChanged.connect(
            lambda v: setattr(self.recording_manager.current, "sampling_rate", v)
        )
        self.ui.epoch_length_input.valueChanged.connect(
            lambda v: setattr(self, "epoch_length", v)
        )
        self.ui.recording_file_button.clicked.connect(self.select_recording_file)
        self.ui.select_label_button.clicked.connect(self.select_label_file)
        self.ui.create_label_button.clicked.connect(self.create_label_file)
        self.ui.manual_scoring_button.clicked.connect(self.manual_scoring)
        self.ui.create_calibration_button.clicked.connect(self.create_calibration_file)
        self.ui.select_calibration_button.clicked.connect(self.select_calibration_file)
        self.ui.load_model_button.clicked.connect(partial(self.load_model, None))
        self.ui.score_all_button.clicked.connect(self.score_recordings)
        self.ui.overwritecheckbox.stateChanged.connect(
            lambda v: setattr(self.scoring, "only_overwrite_undefined", bool(v))
        )
        self.ui.save_confidence_checkbox.stateChanged.connect(
            lambda v: setattr(self.scoring, "save_confidence_scores", bool(v))
        )
        self.ui.bout_length_input.valueChanged.connect(
            lambda v: setattr(self.scoring, "min_bout_length", v)
        )
        self.ui.user_manual_button.clicked.connect(self.show_user_manual)
        self.ui.image_number_input.valueChanged.connect(
            lambda v: setattr(self.training, "epochs_per_img", v)
        )
        self.ui.calibrate_checkbox.stateChanged.connect(
            self.update_training_calibration
        )
        self.ui.train_model_button.clicked.connect(self.train_model)
        self.ui.save_config_button.clicked.connect(self.config.save_config)
        self.ui.export_button.clicked.connect(self.export_recording_list)
        self.ui.import_button.clicked.connect(self.import_recording_list)
        self.ui.default_type_button.toggled.connect(self.model_type_radio_buttons)
        self.ui.reset_emg_params_button.clicked.connect(self.config.reset_emg_filter)
        self.ui.reset_hyperparams_button.clicked.connect(self.config.reset_hyperparams)

        # user input: drag and drop
        self.ui.recording_file_label.installEventFilter(self)
        self.ui.label_file_label.installEventFilter(self)
        self.ui.calibration_file_label.installEventFilter(self)
        self.ui.model_label.installEventFilter(self)

        self.show()

    def model_type_radio_buttons(self, default_selected: bool) -> None:
        """Toggle training default or real-time model

        :param default_selected: whether default option is selected
        """
        self.training.model_type = (
            DEFAULT_MODEL_TYPE if default_selected else REAL_TIME_MODEL_TYPE
        )

    def export_recording_list(self) -> None:
        """Save current list of recordings to file"""
        filename = select_save_location(
            self, "Save list of recordings as", "*" + RECORDING_LIST_FILE_TYPE
        )
        if not filename:
            return
        self.recording_manager.export_to_file(filename)
        self.show_message(f"Saved list of recordings to {filename}")

    def import_recording_list(self):
        """Load list of recordings from file, overwriting current list"""
        filename = select_existing_file(
            self, "Select list of recordings", "*" + RECORDING_LIST_FILE_TYPE
        )
        if not filename:
            return

        self.recording_manager.import_from_file(filename)
        self.show_message(f"Loaded list of recordings from {filename}")

    def eventFilter(self, obj: QObject, event: QEvent) -> bool:
        """Filter mouse events to detect when user drags/drops a file

        :param obj: UI object receiving the event
        :param event: mouse event
        :return: whether to filter (block) the event
        """
        filename = None
        if obj in [
            self.ui.recording_file_label,
            self.ui.label_file_label,
            self.ui.calibration_file_label,
            self.ui.model_label,
        ]:
            event.accept()
            if event.type() == QEvent.Drop:
                urls = event.mimeData().urls()
                if len(urls) == 1:
                    filename = os.path.normpath(urls[0].toLocalFile())

        if filename is None:
            return super().eventFilter(obj, event)
        filename = str(filename)

        _, file_extension = os.path.splitext(filename)

        if obj == self.ui.recording_file_label:
            if file_extension in RECORDING_FILE_TYPES:
                self.recording_manager.current.recording_file = filename
                self.ui.recording_file_label.setText(filename)
        elif obj == self.ui.label_file_label:
            if file_extension == LABEL_FILE_TYPE:
                self.recording_manager.current.label_file = filename
                self.ui.label_file_label.setText(filename)
        elif obj == self.ui.calibration_file_label:
            if file_extension == CALIBRATION_FILE_TYPE:
                self.recording_manager.current.calibration_file = filename
                self.ui.calibration_file_label.setText(filename)
        elif obj == self.ui.model_label:
            self.load_model(filename=filename)

        return super().eventFilter(obj, event)

    def train_model(self) -> None:
        """Train a classification model using the current recordings."""
        model_filename = select_save_location(
            self, "Save classification model file as", "*" + MODEL_FILE_TYPE
        )
        if not model_filename:
            self.show_message("Model training canceled, no filename given")
            return

        # Determine calibration fraction
        if self.training.calibrate:
            calibration_fraction = self.ui.calibration_spinbox.value() / 100
        else:
            calibration_fraction = 0

        # Show progress message
        self.show_message("Training, please wait. See console for progress updates.")
        self.ui.message_area.repaint()
        QApplication.processEvents()

        # Create service and run training
        service = TrainingService(progress_callback=self.show_message)
        result = service.train_model(
            recordings=list(self.recording_manager),
            epoch_length=self.epoch_length,
            epochs_per_img=self.training.epochs_per_img,
            model_type=self.training.model_type,
            calibrate=self.training.calibrate,
            calibration_fraction=calibration_fraction,
            brain_state_set=self.config.brain_state_set,
            emg_filter=self.config.emg_filter,
            hyperparameters=self.config.hyperparameters,
            model_filename=model_filename,
            delete_images=self.config.delete_training_images,
        )

        # Display results
        result.report_to(self.show_message)

    def update_training_calibration(self) -> None:
        """Update choice of whether to calibrate model after training"""
        self.training.calibrate = self.ui.calibrate_checkbox.isChecked()
        self.ui.calibration_spinbox.setEnabled(self.training.calibrate)

    def score_recordings(self) -> None:
        """Score all recordings using the classification model."""
        self.ui.score_all_status.setText("running...")
        self.ui.score_all_status.repaint()
        QApplication.processEvents()

        result = score_recording_list(
            recordings=list(self.recording_manager),
            loaded_model=self.loaded_model,
            epoch_length=self.epoch_length,
            only_overwrite_undefined=self.scoring.only_overwrite_undefined,
            save_confidence_scores=self.scoring.save_confidence_scores,
            min_bout_length=self.scoring.min_bout_length,
            brain_state_set=self.config.brain_state_set,
            emg_filter=self.config.emg_filter,
        )

        # Display results
        result.report_to(self.show_message)
        self.ui.score_all_status.setText("error" if not result.success else "")

    def load_model(self, filename=None) -> None:
        """Load trained classification model from file

        :param filename: model filename, if it's known
        """
        if filename is None:
            filename = select_existing_file(
                self, "Select classification model", "*" + MODEL_FILE_TYPE
            )
            if not filename:
                return

        if not os.path.isfile(filename):
            self.show_message("ERROR: model file does not exist")
            return

        self.show_message("Loading classification model")
        self.ui.message_area.repaint()
        QApplication.processEvents()

        from accusleepy.models import load_model

        try:
            model, epoch_length, epochs_per_img, model_type, brain_states = load_model(
                filename=filename
            )
        except Exception:
            logger.exception("Failed to load %s", filename)
            self.show_message(
                "ERROR: could not load classification model. Check "
                "user manual for instructions on creating this file."
            )
            return

        # make sure only "default" model type is loaded
        if model_type != DEFAULT_MODEL_TYPE:
            self.show_message(
                "ERROR: only 'default'-style models can be used. "
                "'Real-time' models are not supported. "
                "See classification.example_real_time_scoring_function.py "
                "for an example of how to classify brain states in real time."
            )
            return

        self.loaded_model.model = model
        self.loaded_model.epoch_length = epoch_length
        self.loaded_model.epochs_per_img = epochs_per_img

        # warn user if the model's expected epoch length or brain states
        # don't match the current configuration
        config_warnings = check_config_consistency(
            current_brain_states=self.config.brain_state_set.to_output_dict()[
                BRAIN_STATES_KEY
            ],
            model_brain_states=brain_states,
            current_epoch_length=self.epoch_length,
            model_epoch_length=epoch_length,
        )
        if len(config_warnings) > 0:
            for w in config_warnings:
                self.show_message(w)
        else:
            self.show_message(f"Loaded classification model from {filename}")

        self.ui.model_label.setText(filename)

    def load_single_recording(
        self, status_widget: QLabel
    ) -> (np.array, np.array, int | float, bool):
        """Load and preprocess one recording

        This loads one recording, resamples it, and standardizes its length.
        If an error occurs during this process, it is displayed in the
        indicated widget.

        :param status_widget: UI element on which to display error messages
        :return: EEG data, EMG data, sampling rate, process completion
        """
        error_message = check_single_file_inputs(
            self.recording_manager.current, self.epoch_length
        )
        if error_message:
            status_widget.setText(error_message)
            self.show_message(f"ERROR: {error_message}")
            return None, None, None, False

        try:
            eeg, emg = load_recording(self.recording_manager.current.recording_file)
        except Exception:
            logger.exception(
                "Failed to load %s",
                self.recording_manager.current.recording_file,
            )
            status_widget.setText("could not load recording")
            self.show_message(
                "ERROR: could not load recording. "
                "Check user manual for formatting instructions."
            )
            return None, None, None, False

        sampling_rate = self.recording_manager.current.sampling_rate

        eeg, emg, sampling_rate = resample_and_standardize(
            eeg=eeg,
            emg=emg,
            sampling_rate=sampling_rate,
            epoch_length=self.epoch_length,
        )

        return eeg, emg, sampling_rate, True

    def create_calibration_file(self) -> None:
        """Creates a calibration file.

        This loads a recording and its labels, checks that the labels are
        all valid, creates the calibration file, and sets the
        "calibration file" property of the current recording to be the
        newly created file.
        """
        filename = select_save_location(
            self, "Save calibration file as", "*" + CALIBRATION_FILE_TYPE
        )
        if not filename:
            return

        result = create_calibration(
            recording=self.recording_manager.current,
            epoch_length=self.epoch_length,
            brain_state_set=self.config.brain_state_set,
            emg_filter=self.config.emg_filter,
            output_filename=filename,
        )

        # Display results
        result.report_to(self.show_message)
        if not result.success:
            self.ui.calibration_status.setText("error")
        else:
            self.ui.calibration_status.setText("")
            self.recording_manager.current.calibration_file = filename
            self.ui.calibration_file_label.setText(filename)

    def manual_scoring(self) -> None:
        """View the selected recording for manual scoring"""
        # immediately display a status message
        self.ui.manual_scoring_status.setText("loading...")
        self.ui.manual_scoring_status.repaint()
        QApplication.processEvents()

        # load the recording
        eeg, emg, sampling_rate, success = self.load_single_recording(
            self.ui.manual_scoring_status
        )
        if not success:
            return

        # if the labels exist, load them
        # otherwise, create a blank set of labels
        label_file = self.recording_manager.current.label_file
        if os.path.isfile(label_file):
            try:
                labels, confidence_scores = load_labels(label_file)
            except Exception:
                logger.exception("Failed to load %s", label_file)
                self.ui.manual_scoring_status.setText("could not load labels")
                self.show_message(
                    "ERROR: could not load labels. "
                    "Check user manual for formatting instructions."
                )
                return
        else:
            labels = (
                np.ones(int(eeg.size / (sampling_rate * self.epoch_length)))
                * UNDEFINED_LABEL
            ).astype(int)
            # manual scoring will not add a new confidence score column
            # to a label file that does not have one
            confidence_scores = None

        # check that labels are valid and correct minor length mismatches
        labels, confidence_scores, validation_message = validate_and_correct_labels(
            labels=labels,
            confidence_scores=confidence_scores,
            samples_in_recording=eeg.size,
            sampling_rate=sampling_rate,
            epoch_length=self.epoch_length,
            brain_state_set=self.config.brain_state_set,
        )
        if labels is None:
            self.ui.manual_scoring_status.setText("invalid label file")
            self.show_message(f"ERROR: {validation_message}")
            return
        if validation_message:
            self.show_message(f"WARNING: {validation_message}")

        self.show_message(f"Viewing recording {self.recording_manager.current.name}")
        self.ui.manual_scoring_status.setText("file is open")

        # launch the manual scoring window
        manual_scoring_window = ManualScoringWindow(
            eeg=eeg,
            emg=emg,
            label_file=label_file,
            labels=labels,
            confidence_scores=confidence_scores,
            sampling_rate=sampling_rate,
            epoch_length=self.epoch_length,
            emg_filter=self.config.emg_filter,
        )
        manual_scoring_window.setWindowTitle(f"AccuSleePy viewer: {label_file}")
        manual_scoring_window.exec()
        self.ui.manual_scoring_status.setText("")

    def create_label_file(self) -> None:
        """Set the filename for a new label file"""
        filename = select_save_location(
            self,
            "Set filename for label file (nothing will be overwritten yet)",
            "*" + LABEL_FILE_TYPE,
        )
        if filename:
            self.recording_manager.current.label_file = filename
            self.ui.label_file_label.setText(filename)

    def select_label_file(self) -> None:
        """User can select an existing label file"""
        filename = select_existing_file(
            self, "Select label file", "*" + LABEL_FILE_TYPE
        )
        if filename:
            self.recording_manager.current.label_file = filename
            self.ui.label_file_label.setText(filename)

    def select_calibration_file(self) -> None:
        """User can select a calibration file"""
        filename = select_existing_file(
            self, "Select calibration file", "*" + CALIBRATION_FILE_TYPE
        )
        if filename:
            self.recording_manager.current.calibration_file = filename
            self.ui.calibration_file_label.setText(filename)

    def select_recording_file(self) -> None:
        """User can select a recording file"""
        file_filter = f"(*{' *'.join(RECORDING_FILE_TYPES)})"
        filename = select_existing_file(self, "Select recording file", file_filter)
        if filename:
            self.recording_manager.current.recording_file = filename
            self.ui.recording_file_label.setText(filename)

    def show_recording_info(self) -> None:
        """Update the UI to show info for the selected recording"""
        recording = self.recording_manager.current
        self.ui.sampling_rate_input.setValue(recording.sampling_rate)
        self.ui.recording_file_label.setText(recording.recording_file)
        self.ui.label_file_label.setText(recording.label_file)
        self.ui.calibration_file_label.setText(recording.calibration_file)

    def show_message(self, message: str) -> None:
        """Display a new message to the user

        :param message: message to display
        """
        self.messages.append(message)
        if len(self.messages) > MESSAGE_BOX_MAX_DEPTH:
            del self.messages[0]
        self.ui.message_area.setText("\n".join(self.messages))
        # scroll to the bottom
        scrollbar = self.ui.message_area.verticalScrollBar()
        scrollbar.setValue(scrollbar.maximum())

    def select_recording(self, _index: int) -> None:
        """Callback for when a recording is selected"""
        self.show_recording_info()
        self.ui.selected_recording_groupbox.setTitle(
            f"Data / actions for Recording {self.recording_manager.current.name}"
        )

    def add_recording(self) -> None:
        """Add new recording to the list"""
        current_sampling_rate = self.recording_manager.current.sampling_rate
        recording = self.recording_manager.add(sampling_rate=current_sampling_rate)
        self.show_message(f"added Recording {recording.name}")

    def remove_recording(self) -> None:
        """Delete selected recording from the list"""
        self.show_message(self.recording_manager.remove_current())

    def show_user_manual(self) -> None:
        """Show a popup window with the user manual"""
        self.popup = QWidget()
        self.popup_vlayout = QVBoxLayout(self.popup)
        self.guide_textbox = QTextBrowser(self.popup)
        self.popup_vlayout.addWidget(self.guide_textbox)

        url = QUrl.fromLocalFile(MAIN_GUIDE_FILE)
        self.guide_textbox.setSource(url)
        self.guide_textbox.setOpenLinks(False)

        self.popup.setGeometry(QRect(100, 100, 600, 600))
        self.popup.show()


def run_primary_window() -> None:
    logging.basicConfig(
        level=logging.INFO,
        format="%(levelname)s - %(name)s - %(message)s",
    )
    app = QApplication(sys.argv)
    AccuSleepWindow()
    sys.exit(app.exec())


if __name__ == "__main__":
    run_primary_window()



"""Brain state bout length enforcement."""

import re
from dataclasses import dataclass
from operator import attrgetter

import numpy as np


@dataclass
class Bout:
    """Stores information about a brain state bout"""

    length: int  # length, in number of epochs
    start_index: int  # index where bout starts
    end_index: int  # index where bout ends (non-inclusive)
    surrounding_state: int  # brain state on both sides of the bout


def find_last_adjacent_bout(sorted_bouts: list[Bout], bout_index: int) -> int:
    """Find index of last consecutive same-length bout

     When running the post-processing step that enforces a minimum duration
     for brain state bouts, there is a special case when bouts below the
     duration threshold occur consecutively. This function performs a
     recursive search for the index of a bout at the end of such a sequence.
     When initially called, bout_index will always be 0. If, for example, the
     first three bouts in the list are consecutive, the function will return 2.

    :param sorted_bouts: list of brain state bouts, sorted by start time
    :param bout_index: index of the bout in question
    :return: index of the last consecutive same-length bout
    """
    # if we're at the end of the bout list, stop
    if bout_index == len(sorted_bouts) - 1:
        return bout_index

    # if there is an adjacent bout
    if sorted_bouts[bout_index].end_index == sorted_bouts[bout_index + 1].start_index:
        # look for more adjacent bouts using that one as a starting point
        return find_last_adjacent_bout(sorted_bouts, bout_index + 1)
    else:
        return bout_index


def find_short_bouts(
    labels: np.ndarray, min_epochs: int, brain_states: set[int]
) -> list[Bout]:
    """Locate all brain state bouts below a minimum length

    :param labels: brain state labels (digits in the 0-9 range)
    :param min_epochs: minimum number of epochs in a bout
    :param brain_states: set of brain states in the labels
    :return: list of Bout objects
    """
    # convert labels to a string for regex search
    # There is probably a regex that can find all patterns like ab+a
    # without consuming each "a" but I haven't found it :(
    label_string = "".join(labels.astype(str))
    bouts = list()
    for state in brain_states:
        for other_state in brain_states:
            if state == other_state:
                continue
            # get start and end indices of each bout
            expression = (
                f"(?<={other_state}){state}{{1,{min_epochs - 1}}}(?={other_state})"
            )
            matches = re.finditer(expression, label_string)
            spans = [match.span() for match in matches]

            for span in spans:
                bouts.append(
                    Bout(
                        length=span[1] - span[0],
                        start_index=span[0],
                        end_index=span[1],
                        surrounding_state=other_state,
                    )
                )
    return bouts


def enforce_min_bout_length(
    labels: np.ndarray, epoch_length: int | float, min_bout_length: int | float
) -> np.ndarray:
    """Ensure brain state bouts meet the min length requirement

    As a post-processing step for sleep scoring, we can require that any
    bout (continuous period) of a brain state have a minimum duration.
    This function sets any bout shorter than the minimum duration to the
    surrounding brain state (if the states on the left and right sides
    are the same). In the case where there are consecutive short bouts,
    it either creates a transition at the midpoint or removes all short
    bouts, depending on whether the number is even or odd. For example:
    ...AAABABAAA...  -> ...AAAAAAAAA...
    ...AAABABABBB... -> ...AAAAABBBBB...

    :param labels: brain state labels (digits in the 0-9 range)
    :param epoch_length: epoch length, in seconds
    :param min_bout_length: minimum bout length, in seconds
    :return: updated brain state labels
    """
    # if the recording is very short or the minimum bout length
    # is one epoch long, don't change anything
    if labels.size < 3 or epoch_length == min_bout_length:
        return labels

    # get minimum number of epochs in a bout
    min_epochs = int(np.ceil(min_bout_length / epoch_length))
    # get set of states in the labels
    brain_states = set(labels.tolist())

    while True:
        bouts = find_short_bouts(labels, min_epochs, brain_states)
        if len(bouts) == 0:
            break

        # only keep the shortest bouts
        min_length_in_list = np.min([bout.length for bout in bouts])
        bouts = [i for i in bouts if i.length == min_length_in_list]
        # sort by start index
        sorted_bouts = sorted(bouts, key=attrgetter("start_index"))

        while len(sorted_bouts) > 0:
            # get row index of latest adjacent bout (of same length)
            last_adjacent_bout_index = find_last_adjacent_bout(sorted_bouts, 0)
            # if there's an even number of adjacent bouts
            if (last_adjacent_bout_index + 1) % 2 == 0:
                midpoint = sorted_bouts[
                    round((last_adjacent_bout_index + 1) / 2)
                ].start_index
                labels[sorted_bouts[0].start_index : midpoint] = sorted_bouts[
                    0
                ].surrounding_state
                labels[midpoint : sorted_bouts[last_adjacent_bout_index].end_index] = (
                    sorted_bouts[last_adjacent_bout_index].surrounding_state
                )
            else:
                labels[
                    sorted_bouts[0].start_index : sorted_bouts[
                        last_adjacent_bout_index
                    ].end_index
                ] = sorted_bouts[0].surrounding_state

            # delete the bouts we just fixed
            if last_adjacent_bout_index == len(sorted_bouts) - 1:
                sorted_bouts = []
            else:
                sorted_bouts = sorted_bouts[(last_adjacent_bout_index + 1) :]

    return labels



"""Recording list manager"""

from dataclasses import dataclass

from PySide6.QtCore import QObject
from PySide6.QtWidgets import QListWidget, QListWidgetItem

from accusleepy.fileio import Recording, load_recording_list, save_recording_list


@dataclass
class RecordingListItem(Recording):
    """A Recording with an associated QListWidget item for the GUI"""

    widget: QListWidgetItem = None


class RecordingListManager(QObject):
    """Manages the list of recordings and the associated QListWidget"""

    def __init__(self, list_widget: QListWidget, parent: QObject | None = None):
        super().__init__(parent)
        self._widget = list_widget

        # Create initial empty recording (there is always at least one)
        first_recording = RecordingListItem(
            widget=QListWidgetItem("Recording 1", self._widget),
        )
        self._recordings: list[RecordingListItem] = [first_recording]
        self._widget.addItem(first_recording.widget)
        self._widget.setCurrentRow(0)

    @property
    def current(self) -> RecordingListItem:
        """The currently selected recording"""
        return self._recordings[self._widget.currentRow()]

    def add(self, sampling_rate: int | float) -> RecordingListItem:
        """Add a new recording to the list

        :param sampling_rate: sampling rate for the new recording
        :return: the newly created Recording
        """
        new_name = max(r.name for r in self._recordings) + 1

        # Create recording with widget
        recording = RecordingListItem(
            name=new_name,
            sampling_rate=sampling_rate,
            widget=QListWidgetItem(f"Recording {new_name}", self._widget),
        )
        self._recordings.append(recording)

        # Update widget
        self._widget.addItem(recording.widget)
        self._widget.setCurrentRow(len(self._recordings) - 1)

        return recording

    def remove_current(self) -> str:
        """Remove the currently selected recording

        :return: message describing what was removed/reset
        """
        if len(self._recordings) > 1:
            # Remove from list and widget
            index = self._widget.currentRow()
            recording_name = self._recordings[index].name
            self._widget.takeItem(index)
            del self._recordings[index]
            return f"deleted Recording {recording_name}"
        else:
            # Reset the single recording to defaults
            recording_name = self._recordings[0].name
            self._recordings[0] = RecordingListItem(widget=self._recordings[0].widget)
            self._recordings[0].widget.setText(f"Recording {self._recordings[0].name}")
            return f"cleared Recording {recording_name}"

    def export_to_file(self, filename: str) -> None:
        """Save the recording list to a file

        :param filename: path to which the list will be exported
        """
        save_recording_list(filename=filename, recordings=self._recordings)

    def import_from_file(self, filename: str) -> None:
        """Load a recording list from a file, replacing current list

        :param filename: path to load from
        """
        # Block signals while rebuilding the list to avoid triggering
        # selection callbacks with invalid state
        self._widget.blockSignals(True)
        try:
            self._widget.clear()

            # Load recordings and create widgets
            self._recordings = [
                RecordingListItem(
                    **r.__dict__,
                    widget=QListWidgetItem(f"Recording {r.name}", self._widget),
                )
                for r in load_recording_list(filename)
            ]
            for recording in self._recordings:
                self._widget.addItem(recording.widget)
        finally:
            self._widget.blockSignals(False)

        # Select first recording (this will trigger the selection callback)
        self._widget.setCurrentRow(0)

    def __iter__(self):
        return iter(self._recordings)

    def __len__(self):
        return len(self._recordings)

    def __getitem__(self, index: int) -> RecordingListItem:
        return self._recordings[index]



"""Define and organize brain states."""

from dataclasses import dataclass

import numpy as np

BRAIN_STATES_KEY = "brain_states"


@dataclass
class BrainState:
    """Convenience class for a brain state and its attributes"""

    name: str  # friendly name
    digit: int  # number 0-9 - used as keyboard shortcut and in label files
    is_scored: bool  # whether a classification model should score this state
    frequency: int | float  # typical relative frequency, between 0 and 1


class BrainStateSet:
    def __init__(self, brain_states: list[BrainState], undefined_label: int):
        """Initialize set of brain states

        :param brain_states: list of BrainState objects
        :param undefined_label: label for undefined epochs
        """
        self.brain_states = brain_states

        # The user can choose any subset of the digits 0-9 to represent
        # brain states, but not all of them are necessarily intended to be
        # scored by a classifier, and pytorch requires that all input
        # labels are in the 0-n range for training and inference.
        # So, we have to have a distinction between "brain states" (as
        # represented in label files and keyboard inputs) and "classes"
        # (AccuSleep's internal representation).

        # map digits to classes, and vice versa
        self.digit_to_class = {undefined_label: None}
        self.class_to_digit = dict()
        # relative frequencies of each class
        self.mixture_weights = list()

        i = 0
        for brain_state in self.brain_states:
            if brain_state.digit == undefined_label:
                raise ValueError(
                    f"Digit for {brain_state.name} matches 'undefined' label"
                )
            if brain_state.is_scored:
                self.digit_to_class[brain_state.digit] = i
                self.class_to_digit[i] = brain_state.digit
                self.mixture_weights.append(brain_state.frequency)
                i += 1
            else:
                self.digit_to_class[brain_state.digit] = None

        self.n_classes = i

        self.mixture_weights = np.array(self.mixture_weights)
        if np.sum(self.mixture_weights) != 1:
            raise ValueError(
                "Typical frequencies for scored brain states must sum to 1"
            )

    def convert_digit_to_class(self, digits: np.ndarray) -> np.ndarray:
        """Convert array of digits to their corresponding classes

        :param digits: array of digits
        :return: array of classes
        """
        return np.array([self.digit_to_class[i] for i in digits])

    def convert_class_to_digit(self, classes: np.ndarray) -> np.ndarray:
        """Convert array of classes to their corresponding digits

        :param classes: array of classes
        :return: array of digits
        """
        return np.array([self.class_to_digit[i] for i in classes])

    def to_output_dict(self) -> dict:
        """Return dictionary of brain states"""
        return {
            BRAIN_STATES_KEY: [
                {
                    "name": b.name,
                    "digit": b.digit,
                    "is_scored": b.is_scored,
                    "frequency": b.frequency,
                }
                for b in self.brain_states
            ]
        }



"""Neural network model definitions for sleep stage classification."""

from torch import device, flatten, nn
from torch import load as torch_load
from torch import save as torch_save

from accusleepy.brain_state_set import BRAIN_STATES_KEY, BrainStateSet
from accusleepy.constants import (
    IMAGE_HEIGHT,
)
from accusleepy.temperature_scaling import ModelWithTemperature


class SSANN(nn.Module):
    """Small CNN for classifying images"""

    def __init__(self, n_classes: int):
        super().__init__()

        self.pool = nn.MaxPool2d(2, 2)
        self.conv1 = nn.Conv2d(
            in_channels=1, out_channels=8, kernel_size=3, padding="same"
        )
        self.conv2 = nn.Conv2d(
            in_channels=8, out_channels=16, kernel_size=3, padding="same"
        )
        self.conv3 = nn.Conv2d(
            in_channels=16, out_channels=32, kernel_size=3, padding="same"
        )
        self.conv1_bn = nn.BatchNorm2d(8)
        self.conv2_bn = nn.BatchNorm2d(16)
        self.conv3_bn = nn.BatchNorm2d(32)
        self.fc1 = nn.Linear(int(32 * IMAGE_HEIGHT / 8), n_classes)

    def forward(self, x):
        x = x.float()
        x = self.pool(nn.functional.relu(self.conv1_bn(self.conv1(x))))
        x = self.pool(nn.functional.relu(self.conv2_bn(self.conv2(x))))
        x = self.pool(nn.functional.relu(self.conv3_bn(self.conv3(x))))
        x = flatten(x, 1)  # flatten all dimensions except batch
        return self.fc1(x)


def save_model(
    model: SSANN | ModelWithTemperature,
    filename: str,
    epoch_length: int | float,
    epochs_per_img: int,
    model_type: str,
    brain_state_set: BrainStateSet,
    is_calibrated: bool,
) -> None:
    """Save classification model and its metadata

    :param model: classification model
    :param filename: filename
    :param epoch_length: epoch length used when training the model
    :param epochs_per_img: number of epochs in each model input
    :param model_type: default or real-time
    :param brain_state_set: set of brain state options
    :param is_calibrated: whether the model has been calibrated
    """
    state_dict = model.state_dict()
    state_dict.update({"epoch_length": epoch_length})
    state_dict.update({"epochs_per_img": epochs_per_img})
    state_dict.update({"model_type": model_type})
    state_dict.update({"is_calibrated": is_calibrated})
    state_dict.update(
        {BRAIN_STATES_KEY: brain_state_set.to_output_dict()[BRAIN_STATES_KEY]}
    )

    torch_save(state_dict, filename)


def load_model(filename: str) -> tuple[SSANN, int | float, int, str, dict]:
    """Load classification model and its metadata

    :param filename: filename
    :return: model, epoch length used when training the model,
        number of epochs in each model input, model type
        (default or real-time), set of brain state options
        used when training the model
    """
    state_dict = torch_load(filename, weights_only=True, map_location=device("cpu"))
    epoch_length = state_dict.pop("epoch_length")
    epochs_per_img = state_dict.pop("epochs_per_img")
    model_type = state_dict.pop("model_type")
    if "is_calibrated" in state_dict:
        is_calibrated = state_dict.pop("is_calibrated")
    else:
        is_calibrated = False
    brain_states = state_dict.pop(BRAIN_STATES_KEY)
    n_classes = len([b for b in brain_states if b["is_scored"]])

    model = SSANN(n_classes=n_classes)
    if is_calibrated:
        model = ModelWithTemperature(model)
    model.load_state_dict(state_dict)
    return model, epoch_length, epochs_per_img, model_type, brain_states



"""
This is entirely written by the Prerau Lab. Retrieved from this URL in 2025:
https://github.com/preraulab/multitaper_toolbox/blob/master/python/multitaper_spectrogram_python.py
Only minor modifications have been made.
For more information, see their publication:
"Sleep Neurophysiological Dynamics Through the Lens of Multitaper Spectral Analysis"
Michael J. Prerau, Ritchie E. Brown, Matt T. Bianchi, Jeffrey M. Ellenbogen, Patrick L. Purdon
December 7, 2016 : 60-92
DOI: 10.1152/physiol.00062.2015
"""

import math
import timeit
import warnings

import numpy as np


# MULTITAPER SPECTROGRAM #
def spectrogram(
    data,
    fs,
    frequency_range=None,
    time_bandwidth=5,
    num_tapers=None,
    window_params=None,
    min_nfft=0,
    detrend_opt="off",  # this functionality is disabled
    multiprocess=False,
    n_jobs=None,
    weighting="unity",
    plot_on=False,
    return_fig=False,
    clim_scale=True,
    verbose=False,
    xyflip=False,
    ax=None,
):
    """Compute multitaper spectrogram of timeseries data
      Usage:
      mt_spectrogram, stimes, sfreqs = spectrogram(data, fs, frequency_range=None, time_bandwidth=5,
                                                              num_tapers=None, window_params=None, min_nfft=0,
                                                              detrend_opt='linear', multiprocess=False, cpus=False,
                                                              weighting='unity', plot_on=True, return_fig=False,
                                                              clim_scale=True, verbose=True, xyflip=False):
          Arguments:
                  data (1d np.array): time series data -- required
                  fs (float): sampling frequency in Hz  -- required
                  frequency_range (list): 1x2 list - [<min frequency>, <max frequency>] (default: [0 nyquist])
                  time_bandwidth (float): time-half bandwidth product (window duration*half bandwidth of main lobe)
                                          (default: 5 Hz*s)
                  num_tapers (int): number of DPSS tapers to use (default: [will be computed
                                    as floor(2*time_bandwidth - 1)])
                  window_params (list): 1x2 list - [window size (seconds), step size (seconds)] (default: [5 1])
                  detrend_opt (string): detrend data window ('linear' (default), 'constant', 'off')
                                        (Default: 'linear')
                  min_nfft (int): minimum allowable NFFT size, adds zero padding for interpolation (closest 2^x)
                                  (default: 0)
                  multiprocess (bool): Use multiprocessing to compute multitaper spectrogram (default: False)
                  n_jobs (int): Number of cpus to use if multiprocess = True (default: False). Note: if default is left
                              as None and multiprocess = True, the number of cpus used for multiprocessing will be
                              all available - 1.
                  weighting (str): weighting of tapers ('unity' (default), 'eigen', 'adapt');
                  plot_on (bool): plot results (default: True)
                  return_fig (bool): return plotted spectrogram (default: False)
                  clim_scale (bool): automatically scale the colormap on the plotted spectrogram (default: True)
                  verbose (bool): display spectrogram properties (default: True)
                  xyflip (bool): transpose the mt_spectrogram output (default: False)
                  ax (axes): a matplotlib axes to plot the spectrogram on (default: None)
          Returns:
                  mt_spectrogram (TxF np array): spectral power matrix
                  stimes (1xT np array): timepoints (s) in mt_spectrogram
                  sfreqs (1xF np array)L frequency values (Hz) in mt_spectrogram

          Example:
          In this example we create some chirp data and run the multitaper spectrogram on it.
              import numpy as np  # import numpy
              from scipy.signal import chirp  # import chirp generation function
              # Set spectrogram params
              fs = 200  # Sampling Frequency
              frequency_range = [0, 25]  # Limit frequencies from 0 to 25 Hz
              time_bandwidth = 3  # Set time-half bandwidth
              num_tapers = 5  # Set number of tapers (optimal is time_bandwidth*2 - 1)
              window_params = [4, 1]  # Window size is 4s with step size of 1s
              min_nfft = 0  # No minimum nfft
              detrend_opt = 'constant'  # detrend each window by subtracting the average
              multiprocess = True  # use multiprocessing
              cpus = 3  # use 3 cores in multiprocessing
              weighting = 'unity'  # weight each taper at 1
              plot_on = True  # plot spectrogram
              return_fig = False  # do not return plotted spectrogram
              clim_scale = False # don't auto-scale the colormap
              verbose = True  # print extra info
              xyflip = False  # do not transpose spect output matrix

              # Generate sample chirp data
              t = np.arange(1/fs, 600, 1/fs)  # Create 10 min time array from 1/fs to 600 stepping by 1/fs
              f_start = 1  # Set chirp freq range min (Hz)
              f_end = 20  # Set chirp freq range max (Hz)
              data = chirp(t, f_start, t[-1], f_end, 'logarithmic')
              # Compute the multitaper spectrogram
              spect, stimes, sfreqs = spectrogram(data, fs, frequency_range, time_bandwidth, num_tapers,
                                                             window_params, min_nfft, detrend_opt, multiprocess,
                                                             cpus, weighting, plot_on, return_fig, clim_scale,
                                                             verbose, xyflip):

          This code is companion to the paper:
          "Sleep Neurophysiological Dynamics Through the Lens of Multitaper Spectral Analysis"
             Michael J. Prerau, Ritchie E. Brown, Matt T. Bianchi, Jeffrey M. Ellenbogen, Patrick L. Purdon
             December 7, 2016 : 60-92
             DOI: 10.1152/physiol.00062.2015
           which should be cited for academic use of this code.

           A full tutorial on the multitaper spectrogram can be found at: # https://www.sleepEEG.org/multitaper

          Copyright 2021 Michael J. Prerau Laboratory. - https://www.sleepEEG.org
          Authors: Michael J. Prerau, Ph.D., Thomas Possidente, Mingjian He

    __________________________________________________________________________________________________________________
    """
    from scipy.signal.windows import dpss

    #  Process user input
    [
        data,
        fs,
        frequency_range,
        time_bandwidth,
        num_tapers,
        winsize_samples,
        winstep_samples,
        window_start,
        _num_windows,
        nfft,
        detrend_opt,
        plot_on,
        verbose,
    ] = process_input(
        data,
        fs,
        frequency_range,
        time_bandwidth,
        num_tapers,
        window_params,
        min_nfft,
        detrend_opt,
        plot_on,
        verbose,
    )

    # Set up spectrogram parameters
    [window_idxs, stimes, sfreqs, freq_inds] = process_spectrogram_params(
        fs, nfft, frequency_range, window_start, winsize_samples
    )
    # Display spectrogram parameters
    if verbose:
        display_spectrogram_props(
            fs,
            time_bandwidth,
            num_tapers,
            [winsize_samples, winstep_samples],
            frequency_range,
            nfft,
            detrend_opt,
        )

    # Split data into segments and preallocate
    data_segments = data[window_idxs]

    # COMPUTE THE MULTITAPER SPECTROGRAM
    #     STEP 1: Compute DPSS tapers based on desired spectral properties
    #     STEP 2: Multiply the data segment by the DPSS Tapers
    #     STEP 3: Compute the spectrum for each tapered segment
    #     STEP 4: Take the mean of the tapered spectra

    # Compute DPSS tapers (STEP 1)
    dpss_tapers, dpss_eigen = dpss(
        winsize_samples, time_bandwidth, num_tapers, return_ratios=True
    )
    dpss_eigen = np.reshape(dpss_eigen, (num_tapers, 1))

    # pre-compute weights
    if weighting == "eigen":
        wt = dpss_eigen / num_tapers
    elif weighting == "unity":
        wt = np.ones(num_tapers) / num_tapers
        wt = np.reshape(wt, (num_tapers, 1))  # reshape as column vector
    else:
        wt = 0

    tic = timeit.default_timer()  # start timer

    # Set up calc_mts_segment() input arguments
    mts_params = (
        dpss_tapers,
        nfft,
        freq_inds,
        detrend_opt,
        num_tapers,
        dpss_eigen,
        weighting,
        wt,
    )

    mt_spectrogram = np.apply_along_axis(
        calc_mts_segment, 1, data_segments, *mts_params
    )

    # Compute one-sided PSD spectrum
    mt_spectrogram = mt_spectrogram.T
    dc_select = np.where(sfreqs == 0)[0]
    nyquist_select = np.where(sfreqs == fs / 2)[0]
    select = np.setdiff1d(
        np.arange(0, len(sfreqs)), np.concatenate((dc_select, nyquist_select))
    )

    mt_spectrogram = (
        np.vstack(
            [
                mt_spectrogram[dc_select, :],
                2 * mt_spectrogram[select, :],
                mt_spectrogram[nyquist_select, :],
            ]
        )
        / fs
    )

    # Flip if requested
    if xyflip:
        mt_spectrogram = mt_spectrogram.T

    # End timer and get elapsed compute time
    toc = timeit.default_timer()
    if verbose:
        print("\n Multitaper compute time: " + "%.2f" % (toc - tic) + " seconds")

    if np.all(mt_spectrogram.flatten() == 0):
        print("\n Data was all zeros, no output")

    return mt_spectrogram, stimes, sfreqs


# Helper Functions #


# Process User Inputs #
def process_input(
    data,
    fs,
    frequency_range=None,
    time_bandwidth=5,
    num_tapers=None,
    window_params=None,
    min_nfft=0,
    detrend_opt="linear",
    plot_on=True,
    verbose=True,
):
    """Helper function to process spectrogram() arguments
    Arguments:
            data (1d np.array): time series data-- required
            fs (float): sampling frequency in Hz  -- required
            frequency_range (list): 1x2 list - [<min frequency>, <max frequency>] (default: [0 nyquist])
            time_bandwidth (float): time-half bandwidth product (window duration*half bandwidth of main lobe)
                                    (default: 5 Hz*s)
            num_tapers (int): number of DPSS tapers to use (default: None [will be computed
                              as floor(2*time_bandwidth - 1)])
            window_params (list): 1x2 list - [window size (seconds), step size (seconds)] (default: [5 1])
            min_nfft (int): minimum allowable NFFT size, adds zero padding for interpolation (closest 2^x)
                            (default: 0)
            detrend_opt (string): detrend data window ('linear' (default), 'constant', 'off')
                                  (Default: 'linear')
            plot_on (True): plot results (default: True)
            verbose (True): display spectrogram properties (default: true)
    Returns:
            data (1d np.array): same as input
            fs (float): same as input
            frequency_range (list): same as input or calculated from fs if not given
            time_bandwidth (float): same as input or default if not given
            num_tapers (int): same as input or calculated from time_bandwidth if not given
            winsize_samples (int): number of samples in single time window
            winstep_samples (int): number of samples in a single window step
            window_start (1xm np.array): array of timestamps representing the beginning time for each window
            num_windows (int): number of windows in the data
            nfft (int): length of signal to calculate fft on
            detrend_opt ('string'): same as input or default if not given
            plot_on (bool): same as input
            verbose (bool): same as input
    """

    # Make sure data is 1 dimensional np array
    if len(data.shape) != 1:
        if (len(data.shape) == 2) & (
            data.shape[1] == 1
        ):  # if it's 2d, but can be transferred to 1d, do so
            data = np.ravel(data[:, 0])
        elif (len(data.shape) == 2) & (
            data.shape[0] == 1
        ):  # if it's 2d, but can be transferred to 1d, do so
            data = np.ravel(data.T[:, 0])
        else:
            raise TypeError(
                "Input data is the incorrect dimensions. Should be a 1d array with shape (n,) where n is \
                            the number of data points. Instead data shape was "
                + str(data.shape)
            )

    # Set frequency range if not provided
    if frequency_range is None:
        frequency_range = [0, fs / 2]

    # Set detrending method
    detrend_opt = detrend_opt.lower()
    if detrend_opt != "linear":
        if detrend_opt in ["const", "constant"]:
            detrend_opt = "constant"
        elif detrend_opt in ["none", "false", "off"]:
            detrend_opt = "off"
        else:
            raise ValueError(
                "'"
                + str(detrend_opt)
                + "' is not a valid argument for detrend_opt. The choices "
                + "are: 'constant', 'linear', or 'off'."
            )
    # Check if frequency range is valid
    if frequency_range[1] > fs / 2:
        frequency_range[1] = fs / 2
        warnings.warn(
            "Upper frequency range greater than Nyquist, setting range to ["
            + str(frequency_range[0])
            + ", "
            + str(frequency_range[1])
            + "]",
            stacklevel=2,
        )

    # Set number of tapers if none provided
    if num_tapers is None:
        num_tapers = math.floor(2 * time_bandwidth) - 1

    # Warn if number of tapers is suboptimal
    if num_tapers != math.floor(2 * time_bandwidth) - 1:
        warnings.warn(
            "Number of tapers is optimal at floor(2*TW) - 1. consider using "
            + str(math.floor(2 * time_bandwidth) - 1),
            stacklevel=2,
        )

    # If no window params provided, set to defaults
    if window_params is None:
        window_params = [5, 1]

    # Check if window size is valid, fix if not
    if window_params[0] * fs % 1 != 0:
        winsize_samples = round(window_params[0] * fs)
        warnings.warn(
            "Window size is not divisible by sampling frequency. Adjusting window size to "
            + str(winsize_samples / fs)
            + " seconds",
            stacklevel=2,
        )
    else:
        winsize_samples = window_params[0] * fs

    # Check if window step is valid, fix if not
    if window_params[1] * fs % 1 != 0:
        winstep_samples = round(window_params[1] * fs)
        warnings.warn(
            "Window step size is not divisible by sampling frequency. Adjusting window step size to "
            + str(winstep_samples / fs)
            + " seconds",
            stacklevel=2,
        )
    else:
        winstep_samples = window_params[1] * fs

    # Get total data length
    len_data = len(data)

    # Check if length of data is smaller than window (bad)
    if len_data < winsize_samples:
        raise ValueError(
            "\nData length ("
            + str(len_data)
            + ") is shorter than window size ("
            + str(winsize_samples)
            + "). Either increase data length or decrease window size."
        )

    # Find window start indices and num of windows
    window_start = np.arange(0, len_data - winsize_samples + 1, winstep_samples)
    num_windows = len(window_start)

    # Get num points in FFT
    if min_nfft == 0:  # avoid divide by zero error in np.log2(0)
        nfft = max(2 ** math.ceil(np.log2(abs(winsize_samples))), winsize_samples)
    else:
        nfft = max(
            max(2 ** math.ceil(np.log2(abs(winsize_samples))), winsize_samples),
            2 ** math.ceil(np.log2(abs(min_nfft))),
        )

    return [
        data,
        fs,
        frequency_range,
        time_bandwidth,
        num_tapers,
        int(winsize_samples),
        int(winstep_samples),
        window_start,
        num_windows,
        nfft,
        detrend_opt,
        plot_on,
        verbose,
    ]


# PROCESS THE SPECTROGRAM PARAMETERS #
def process_spectrogram_params(fs, nfft, frequency_range, window_start, datawin_size):
    """Helper function to create frequency vector and window indices
    Arguments:
         fs (float): sampling frequency in Hz  -- required
         nfft (int): length of signal to calculate fft on -- required
         frequency_range (list): 1x2 list - [<min frequency>, <max frequency>] -- required
         window_start (1xm np array): array of timestamps representing the beginning time for each
                                      window -- required
         datawin_size (float): seconds in one window -- required
    Returns:
        window_idxs (nxm np array): indices of timestamps for each window
                                    (nxm where n=number of windows and m=datawin_size)
        stimes (1xt np array): array of times for the center of the spectral bins
        sfreqs (1xf np array): array of frequency bins for the spectrogram
        freq_inds (1d np array): boolean array of which frequencies are being analyzed in
                                  an array of frequencies from 0 to fs with steps of fs/nfft
    """

    # create frequency vector
    df = fs / nfft
    sfreqs = np.arange(0, fs, df)

    # Get frequencies for given frequency range
    freq_inds = (sfreqs >= frequency_range[0]) & (sfreqs <= frequency_range[1])
    sfreqs = sfreqs[freq_inds]

    # Compute times in the middle of each spectrum
    window_middle_samples = window_start + round(datawin_size / 2)
    stimes = window_middle_samples / fs

    # Get indexes for each window
    window_idxs = np.atleast_2d(window_start).T + np.arange(0, datawin_size, 1)
    window_idxs = window_idxs.astype(int)

    return [window_idxs, stimes, sfreqs, freq_inds]


# DISPLAY SPECTROGRAM PROPERTIES
def display_spectrogram_props(
    fs,
    time_bandwidth,
    num_tapers,
    data_window_params,
    frequency_range,
    nfft,
    detrend_opt,
):
    """Prints spectrogram properties
    Arguments:
        fs (float): sampling frequency in Hz  -- required
        time_bandwidth (float): time-half bandwidth product (window duration*1/2*frequency_resolution) -- required
        num_tapers (int): number of DPSS tapers to use -- required
        data_window_params (list): 1x2 list - [window length(s), window step size(s)] -- required
        frequency_range (list): 1x2 list - [<min frequency>, <max frequency>] -- required
        nfft(float): number of fast fourier transform samples -- required
        detrend_opt (str): detrend data window ('linear' (default), 'constant', 'off') -- required
    Returns:
        This function does not return anything
    """

    data_window_params = np.asarray(data_window_params) / fs

    # Print spectrogram properties
    print("Multitaper Spectrogram Properties: ")
    print(
        "     Spectral Resolution: "
        + str(2 * time_bandwidth / data_window_params[0])
        + "Hz"
    )
    print("     Window Length: " + str(data_window_params[0]) + "s")
    print("     Window Step: " + str(data_window_params[1]) + "s")
    print("     Time Half-Bandwidth Product: " + str(time_bandwidth))
    print("     Number of Tapers: " + str(num_tapers))
    print(
        "     Frequency Range: "
        + str(frequency_range[0])
        + "-"
        + str(frequency_range[1])
        + "Hz"
    )
    print("     NFFT: " + str(nfft))
    print("     Detrend: " + detrend_opt + "\n")


# NANPOW2DB
def nanpow2db(y):
    """Power to dB conversion, setting bad values to nans
    Arguments:
        y (float or array-like): power
    Returns:
        ydB (float or np array): inputs converted to dB with 0s and negatives resulting in nans
    """

    if isinstance(y, int | float):
        if y == 0:
            return np.nan
        else:
            ydB = 10 * np.log10(y)
    else:
        if isinstance(y, list):  # if list, turn into array
            y = np.asarray(y)
        y = y.astype(float)  # make sure it's a float array so we can put nans in it
        y[y == 0] = np.nan
        ydB = 10 * np.log10(y)

    return ydB


# CALCULATE MULTITAPER SPECTRUM ON SINGLE SEGMENT
def calc_mts_segment(
    data_segment,
    dpss_tapers,
    nfft,
    freq_inds,
    detrend_opt,
    num_tapers,
    dpss_eigen,
    weighting,
    wt,
):
    """Helper function to calculate the multitaper spectrum of a single segment of data
    Arguments:
        data_segment (1d np.array): One window worth of time-series data -- required
        dpss_tapers (2d np.array): Parameters for the DPSS tapers to be used.
                                   Dimensions are (num_tapers, winsize_samples) -- required
        nfft (int): length of signal to calculate fft on -- required
        freq_inds (1d np array): boolean array of which frequencies are being analyzed in
                                  an array of frequencies from 0 to fs with steps of fs/nfft
        detrend_opt (str): detrend data window ('linear' (default), 'constant', 'off')
        num_tapers (int): number of tapers being used
        dpss_eigen (np array):
        weighting (str):
        wt (int or np array):
    Returns:
        mt_spectrum (1d np.array): spectral power for single window
    """

    # If segment has all zeros, return vector of zeros
    if all(data_segment == 0):
        ret = np.empty(sum(freq_inds))
        ret.fill(0)
        return ret

    if any(np.isnan(data_segment)):
        ret = np.empty(sum(freq_inds))
        ret.fill(np.nan)
        return ret

    # # Option to detrend data to remove low frequency DC component
    # if detrend_opt != "off":
    #     data_segment = detrend(data_segment, type=detrend_opt)

    # Multiply data by dpss tapers (STEP 2)
    tapered_data = data_segment[:, np.newaxis] * dpss_tapers.T

    # Compute the FFT (STEP 3)
    fft_data = np.fft.fft(tapered_data, nfft, axis=0)

    # Compute the weighted mean spectral power across tapers (STEP 4)
    spower = np.power(np.imag(fft_data), 2) + np.power(np.real(fft_data), 2)
    if weighting == "adapt":
        # adaptive weights - for colored noise spectrum (Percival & Walden p368-370)
        tpower = np.dot(np.transpose(data_segment), (data_segment / len(data_segment)))
        spower_iter = np.mean(spower[:, 0:2], 1)
        spower_iter = spower_iter[:, np.newaxis]
        a = (1 - dpss_eigen) * tpower
        for _i in range(3):  # 3 iterations only
            # Calc the MSE weights
            b = np.dot(spower_iter, np.ones((1, num_tapers))) / (
                (np.dot(spower_iter, np.transpose(dpss_eigen)))
                + (np.ones((nfft, 1)) * np.transpose(a))
            )
            # Calc new spectral estimate
            wk = (b**2) * np.dot(np.ones((nfft, 1)), np.transpose(dpss_eigen))
            spower_iter = np.sum((np.transpose(wk) * np.transpose(spower)), 0) / np.sum(
                wk, 1
            )
            spower_iter = spower_iter[:, np.newaxis]

        mt_spectrum = np.squeeze(spower_iter)

    else:
        # eigenvalue or uniform weights
        mt_spectrum = np.dot(spower, wt)
        mt_spectrum = np.reshape(mt_spectrum, nfft)  # reshape to 1D

    return mt_spectrum[freq_inds]
